#!/usr/bin/env python3
#
#  radio_station_data.py
"""
Data for songs that play on the in game radio.
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# TODO: generes (from tags in sounddb)
# TODO: check in-game capitalisation of station names

# stdlib
from collections.abc import Sequence

# this package
from cp2077_extractor.track import Track

__all__ = [
		"body_heat_radio",
		"dark_star",
		"growl_fm",
		"misc",
		"morro_rock_radio",
		"night_fm",
		"pacific_dreams",
		"principales",
		"radio_jingle_ids",
		"radio_pebkac",
		"radio_stations",
		"radio_vexelstrom",
		"ritual_fm",
		"royal_blue_radio",
		"samizdat_radio",
		"the_dirge",
		]

# TODO: read from JSON file

# Other uses of the songs (mainly in quests).
HANGOUT_QUEST = "Hangout"
RACES = "Races"
MUS_OW_PHONE = "mus_ow_phone"
SOMEWHAT_DAMAGED_FLASHBACK = "mus_q305_brooklyn_vinyl_01"
SOMEWHAT_DAMAGED_FLASHBACK_REED = "mus_q305_brooklyn_vinyl_reed"
CLOUDS_MUSIC = "mus_q105_dollhouse"
DARK_MATTER = "mus_sq017_dark_matter_club_user_friendly"  # Dark Matter during Kerry quest
JOHNNY_KERRY = "mus_q101_js_and_kerry_source"  # Love Like Fire quest Johnny & Kerry

# TODO: or is it 88.3, as in the credits data?
#: 88.9 Pacific Dreams
pacific_dreams: list[Track] = [
		Track("Quantum Lovers", "Isometric Air", 253367813, "Brian Aspey", "Brian Aspey"),
		Track("Quantum Lovers", "Practical Heart", 910852846, "Brian Aspey", "Brian Aspey"),
		Track(
				"Quantum Lovers",
				"Real Window",
				1005242856,
				"Brian Aspey",
				"Brian Aspey", {895217420: HANGOUT_QUEST}
				),
		Track(
				"Pacific Avenue",
				"Antagonistic",
				190841656,
				"Sebastian Robertson and Chris Cardena",
				"Chris Cardena & Sebastian Robertson",
				{871197545: HANGOUT_QUEST},
				),
		Track(
				"Jänsens",
				"Simple Pleasures",
				803750564,
				"Elena Charbila; Dimitris Mann",
				"Kid Moxie", {880346021: HANGOUT_QUEST}
				),
		Track("Flatlander Woman", "Lithium", 770891476, "SLG", "SLG"),
		Track("Flatlander Woman", "Slag", 1044871022, "SLG", "SLG"),
		Track(
				"Muchomorr",
				"Chodze",
				1066343217,
				"Mchy i Porosty",
				"Mchy i Porosty", {443457920: "Parade and Wako's briefing"}
				),
		Track("Lick Switch", "Midnight Eye", 928920080, "Earth Trax", "Earth Trax"),
		Track("Lick Switch", "Blurred", 285910393, "Earth Trax", "Earth Trax"),
		Track("Lick Switch", "The Other Room", 869646378, "Earth Trax", "Earth Trax"),
		Track("Sonoris Causa", "La Stessa Causa", 621406329, "Eltron", "Eltron"),
		Track("Left Unsaid", "Retrogenesis", 562668759, "Private Press", "Private Press"),
		Track("Talk to Us", "Miami Suicide", 400079366, "Chino", "Chino"),
		Track("Talk to Us", "Slippery Stabs", 1056620275, "Chino", "Chino"),
		Track("Wormview", "Ashes and Diamonds", 95368841, "Hatti Vatti", "Hatti Vatti"),
		Track("Mona Mitchell", "Ice Maddox", 1017413855, "FOQL", "FOQL"),
		]

#: 89.3 Radio Vexelstrom
radio_vexelstrom: list[Track] = [
		Track("The Cartesian Duelists", "Resist and Disorder", 308480891, "Jason Charles Miller", "Rezodrone"),
		Track("The Cartesian Duelists", "Kill the Messenger", 525137074, "Jason Charles Miller", "Rezodrone"),
		Track("Slavoj McAllister", "Makes Me Feel Better", 66952680, "Kevin Hastings", "OnenO"),
		Track("Keine", "Dead Pilot", 409198739, "Sebastian Robertson; Daniel Davies", "SRDD (Sebastian Daniel)"),
		Track("Keine", "Come Close", 303931745, "Sebastian Robertson; Daniel Davies", "SRDD (Sebastian Daniel)"),
		Track("Upgrade", "Black Terminal", 1065746543, "Bret Autrey; Danny Cocke", "Black Terminal"),
		Track("Alexei Brayko", "Reaktion", 49937168, "Jason Charles Miller", "Rezodrone"),
		Track("Ego Affliction", "With Her", 257784898, "Steven Richard Davis", "Steven Richard Davis"),
		Track("Den of Degenerates", "Never Stop Me", 811249141, "Steven Richard Davis", "Steven Richard Davis"),
		Track("The Red Glare", "Violence", 588422768, "Kristina Michelle Olson", "Le Destroy"),
		Track("The Red Glare", "Pain", 365987976, "Kristina Michelle Olson", "Le Destroy"),
		Track("Homeschool Dropouts", "Night City Aliens", 86209634, "The Armed", "The Armed"),
		Track("Tainted Overlord", "A Caça", 420329756, "Deafkids", "Deafkids", {640742970: RACES}),
		Track("Tainted Overlord", "Selva Pulsátil", 50379518, "Deafkids", "Deafkids"),
		Track(
				"N1v3Z",
				"Pig Dinner",
				874085468,
				"Ho99o9, N8NOFACE",
				"Ho99o9, N8NOFACE",
				),  # Not listed in credits data
		]

# TODO: check against credits
#: 89.7 Growl FM
growl_fm: list[Track] = [
		Track(
				"Coeur Noir",
				"Let The Stars Die",
				444314240,
				'Jeremy Ghanassia, Maxime "Paka" Marin',
				other_uses={959192573: SOMEWHAT_DAMAGED_FLASHBACK_REED}
				),
		Track(
				"Red Dead Roadkill",
				"Flatline",
				694001176,
				"Julien Röttger",
				"Red Dead Roadkill, Dee Wolf", {896635379: MUS_OW_PHONE}
				),
		Track(
				"Spirit Machines",
				"Candy Shell",
				945083955,
				"Jessica McCombs, Dave Crespo, Michael Collins, Sergio Marticorena"
				),
		Track("NoWorld", "Do or Die", 807725976, "NoWorld"),
		Track("Entolim", "LIT", 265573485, "Entolim"),
		Track("Skin on Flesh", "El Tiempo", 951381637, "Laura Jiménez Alvarez, Michiel Sybers"),
		Track(
				"Frost, Justtjokay, Dubbygotbars, Knyvez",
				"Killshot",
				116162799,
				"Frost, Justtjokay, Dubbygotbars, Knyvez",
				other_uses={316992516: MUS_OW_PHONE, 608799802: SOMEWHAT_DAMAGED_FLASHBACK}
				),
		Track(
				"Haru Nemuri",
				"さまよえるままゆけ (Samayoeru mama yuke) / Let It Go As If You Wander",
				214475991,
				"春ねむり (Haru Nemuri)",
				"春ねむり (Haru Nemuri)"
				),
		Track("Kiba", "Slipstream", 581037122, "Kiba"),
		Track("Aleyna Moon, Shrinjay Ghosh", "FUMES", 252703137, "Aleyna Moon, Shrinjay Ghosh"),
		Track(
				"D.O.H. Dollahz Ova Hoez",
				"Look Through My Kiroshis (The Solo Life)",
				606235279,
				"D.O.H. Dollahz Ova Hoez",
				other_uses={328723041: MUS_OW_PHONE, 1009571067: SOMEWHAT_DAMAGED_FLASHBACK}
				),
		Track("St. Aurora", "Going to Heaven", 816183854, "St. Aurora", other_uses={254921013: MUS_OW_PHONE}),
		Track(
				"Thai McGrath (ft. JustCosplaySings)",
				"Afterlife",
				900442762,
				"Thai McGrath; JustCosplaySings",
				"Thai McGrath; JustCosplaySings", {740594675: MUS_OW_PHONE}
				),
		]

#: 91.9 Royal Blue Radio
royal_blue_radio: list[Track] = [
		Track("Miles Davis", "Black Satin/What If/Agharta Prelude Dub", 448263054, "Miles Davis; Bill Laswell"),
		Track("Miles Davis", "Bitches Brew", 915057747, "Miles Davis"),
		Track(
				"Miles Davis",
				"Générique (from Ascenseur pour L'échafaud)",
				143859087,
				"Miles Davis",
				other_uses={987597831: HANGOUT_QUEST}
				),
		Track("John Coltrane", "Impressions", 230077180, "John Coltrane"),
		Track(
				"Charles Mingus",
				"Solo Dancer",
				1014236840,
				"Charles Mingus",
				),  # From the album The Black Saint and the Sinner Lady
		Track("Dexter Gordon", "Laura", 419089334, "David Raksin"),
		Track(
				"Chet Baker",
				"You Don't Know What Love Is",
				400388835,
				"Gene Depaul; Don Raye",
				other_uses={187092121: HANGOUT_QUEST, 397320180: "Konpeki Plaza Braindance"}
				),
		Track("Thelonious Monk", "Round Midnight", 487426870, "Thelonious Monk; Bernie Hanighen; Cootie Williams"),
		Track("Trio of Doom", "Dark Prince", 449864253, "John Mclaughlin", other_uses={395786567: RACES}),
		]

#: 92.9 Night FM
night_fm: list[Track] = [
		Track("Perilous Futur", "Dirty Roses", 664992914, "Kevin Hastings", "OnenO"),
		Track(
				"The Unresolved",
				"Worlds",
				188143271,
				"Sebastian Robertson, Daniel Davies",
				"SRDD (Sebastian Daniel)"
				),
		Track("The Unresolved", 'X', 52108015, "Sebastian Robertson; Daniel Davies", "SRDD (Sebastian Daniel)"),
		Track("Doctor Berserk", "Maniak", 592937823, "Fabian Velazquez", "Picasso"),
		Track("Generating Dependencies", "Me Machine", 621649083, "Kuba Sojka", "Poly Face"),
		Track("Lick Switch", "Like a Miracle", 791114400, "Earth Trax", "Earth Trax"),
		Track("Kings of Collapse", "Run", 649255606, "Steven Richard Davis", "Steven Richard Davis"),
		Track("Reviscerator", "Glitched Revelation", 521364565, "Procesor Plus", "Procesor Plus"),
		Track("Reviscerator", "Yellow Box", 884931314, "Procesor Plus", "Procesor Plus"),
		Track("The Bait", "Kill Kill", 84073683, "Kristina Michelle Olson", "Le Destroy"),
		Track(
				"Ashes Potts",
				"Flying Heads",
				710361138,
				"Ivan Iusco; Elena Charbila",
				"Kid Moxie; Ivan Iusco; Elena Charbila"
				),
		Track("Yards of the Moon", "Volcano the Sailor", 783271344, "Połoz", "Połoz"),
		Track("Cyber Coorayber", "Brain-Damaged", 239254618, "Nikola Nikita Jeremic", "Nikola Nikita Jeremic"),
		]

#: 95.2 Samizdat Radio
samizdat_radio: list[Track] = [
		Track(
				"Bara Nova",
				"Piling in my Head",
				185138126,
				"Nina Kraviz",
				"Nina Kraviz", {892337963: CLOUDS_MUSIC}
				),
		Track(
				"Bara Nova",
				"Delirium 2",
				625219838,
				"Nina Kraviz",
				"Nina Kraviz",
				{900420316: "mus_sq027_panam_csex_01_pre_sex", 439311782: "mus_sq027_panam_csex_01_pre_sex"}
				),
		Track("Bara Nova", "Harm Sweaty Pit", 928983117, "Nina Kraviz", "Nina Kraviz", {655028819: CLOUDS_MUSIC}),
		Track(
				"Bara Nova",
				"My Lullaby for You",
				740337697,
				"Nina Kraviz",
				"Nina Kraviz",
				{174026915: CLOUDS_MUSIC},
				),
		Track("Bara Nova", "Surprise Me, I'm Surprised Today", 822859958, "Nina Kraviz", "Nina Kraviz"),
		]

#: 96.1 Ritual FM
ritual_fm: list[Track] = [
		Track("V3RM1N", "Finis", 323048169, "Piotr Maciejewski (Drivealone)", "Piotr Maciejewski (Drivealone)"),
		Track("Dread Soul", "The Accursed", 969984534, "Antre", "Antre"),
		Track(
				"Bacillus",
				"March 30",
				643009039,
				"Tomb Mold",
				"Tomb Mold",
				),  # Also called Adaptive Manipulator outside the game.
		Track("Forlorn Scourge", "Acid Breather", 685504893, "Mastiff", "Mastiff"),
		Track("Nuclear Aura", "Witches of the Harz Mountains", 670708886, "Marcin Rybicki", "Marcin Rybicki"),
		Track("Weles", "The Loop", 381506730, "Deszcz", "Deszcz"),
		Track("Hysteria", "Scrum", 516709760, "Totenmesse", "Totenmesse"),
		Track("Inferno Corps", "Fuelled by Poison", 532284261, "Antigama", "Antigama", {544955296: JOHNNY_KERRY}),
		Track("Inferno Corps", "Kevin", 606018074, "Antigama", "Antigama"),
		Track("heXXXer", "Future Drugs", 620624990, "Entropia", "Entropia"),  # Or is the artist Mord'A'Stigmata?
		Track("Wydech", "Żurawie", 101839488, "Ugory", "Ugory"),
		Track(
				"Fist of Satan",
				"Abandoned Land",
				957306104,
				"Artur Rumiński; Haldor Grunberg",
				"Artur Rumiński; Haldor Grunberg"
				),
		Track(
				"Fist of Satan",
				"Black Concrete",
				114092930,
				"Artur Rumiński; Haldor Grunberg",
				"Artur Rumiński; Haldor Grunberg"
				),
		Track("Shattered Void", "I Won't Let You Go", 643641585, "Converge", "Converge"),
		]

#: 98.7 Body Heat Radio
body_heat_radio: list[Track] = [
		Track(
				"Clockwork Venus",
				"BM",
				1017973036,
				"SOPHIE; Shygirl; Sega Bodega; Kai Whiston",
				"SOPHIE; Shygirl",
				),  # Also called SLIME in the game credits and outside the game.
		Track("Neon Haze", "Circus Minimus", 903009003, "Jazelle Rodriguez; Deryk Mitchell; Erika Nuri", "Jvzel"),
		Track(
				"Window Weather",
				"Major Crimes",
				688260506,
				"Jake Duzsik; John Famiglietti; Lars Stalfors",
				"HEALTH",
				{41444121: DARK_MATTER}  # Also for Shy Delamain quest
				),
		Track("Artemis Delta", "Night City", 862539357, "Arielle Sitrick; Bill Burke", "R E L"),
		Track(
				"Hallie Coggins",
				"I Really Want To Stay At Your House",
				717186759,
				"Rosa Walton",
				"Rosa Walton (Let's Eat Grandma)",
				{670626837: HANGOUT_QUEST, 881817538: "mus_q105_fingers_01_source", 890846657: DARK_MATTER}
				),
		Track(
				"Point Break Candy",
				"Hole In The Sun",
				698612311,
				"Raney Shockne",
				"Raney Shockne feat. COS, Conway"
				),
		# TODO: Track("American Medical Association", "Bliind", 0, "Raney Shockne", "Raney Shockne"),
		Track(
				"Trash Generation",
				"History",
				941358832,
				"Gazelle Twin",
				"Gazelle Twin", {986982905: "mus_q000_corpo_standoff_source"}
				),
		Track("Lizzy Wizzy", "4ÆM", 5020242, "Grimes", "Grimes"),
		Track(
				"Lizzy Wizzy",
				"Delicate Weapon",
				726229946,
				"Grimes",
				"Grimes", {670923974: "Phantom Liberty Black Sapphire performance"}
				),
		Track("Us Cracks", "Ponpon Shit", 259328250, "Yuki Kawamura", "Namakopuri", {180075549: DARK_MATTER}),
		Track(
				"Us Cracks",
				"User Friendly",
				503548262,
				"Katarzyna Kraińska; Yuki Kawamura; Kaito Sakuma",
				"Namakopuri",
				{386020430: DARK_MATTER},
				),  # TODO: does this play on this station?
		Track(
				"Us Cracks feat. Kerry Eurodyne",
				"Off The Leash",
				235217575,
				"Kaito Sakuma; Yuki Kawamura; Tomas Shimizu (American Dream Express)",
				"Namakopuri & Damian Ukeje",
				{988042201: DARK_MATTER},
				),
		Track("IBDY", "Crustpunk", 949985077, "Rat Boy", "Rat Boy", {675780732: RACES}),
		Track("IBDY", "Here's a Thought", 575003013, "Rat Boy", "Rat Boy", {825809386: JOHNNY_KERRY}),
		]

#: 99.9 Impulse
impulse: list[Track] = [
		# TODO: songs list.
		Track(
				"Mr. Kipper",
				"DJ Set",
				424884909,
				),  # mus_radio_14_impulse_djset, also mus_radio_14impls_djset_pyramid_edit
		# Track("Mr. Kipper", "Walk Of Shame", 0, "Idris Elba", "Idris Elba"),
		# Track("Private Press", "Dreamy", 0, "Adam Brocki; Jan Wóycicki"),
		# Track("Private Press", "DEEEEEEP", 0, "Adam Brocki; Jan Wóycicki"),
		# Track("Private Press", "Black Labyrinth", 0, "Adam Brocki; Jan Wóycicki"),
		# Track("Private Press", "Feed Your Soul", 0, "Adam Brocki; Jan Wóycicki"),
		# Track("Private Press", "Liquid Disco", 0, "Adam Brocki; Jan Wóycicki"),
		# Track("Private Press", "Sleepy Dust", 0, "Adam Brocki; Jan Wóycicki"),
		# Track("Private Press", "Sparkling Frequency", 0, "Adam Brocki; Jan Wóycicki"),
		# Track("Private Press", "Undeniably Changes", 0, "Adam Brocki; Jan Wóycicki"),
		# Track("Private Press", "Void", 0, "Adam Brocki; Jan Wóycicki"),
		# Track("Private Press", "Woozee", 0, "Adam Brocki; Jan Wóycicki"),
		# Track("Private Press", "GREEEEEENHOUSE", 0, "Adam Brocki; Jan Wóycicki"),
		# Track("Private Press", "BESTPL", 0, "Adam Brocki; Jan Wóycicki"),
		# Track("Private Press", "ENERGEEHOUSE", 0, "Adam Brocki; Jan Wóycicki"),
		]

# TODO: or is it 101.0 like in the credits data
#: 101.9 The Dirge
the_dirge: list[Track] = [
		Track(
				"Kill Trigger feat. Paul Senai, KraKow",
				"The God Machines",
				660406395,
				"Sebastian Robertson; Daniel Davies; Tristan Calder; Brandon Hale; Michael Garcia",
				"Sebastian Robertson, Kill The Computer & Indijinouz"
				),
		Track(
				"NC3",
				"Blouses Blue",
				777673336,
				"Konrad Abramowicz",
				"Konrad OldMoney feat. Cidro Onetoo, Perry Porter"
				),
		Track("Young Kenny", "Problem Kids", 422374976, "Konrad Abramowicz", "Konrad OldMoney feat. Taelor Yung"),
		Track("Droox", "Bigger Man", 305730733, "Konrad Abramowicz", "Konrad OldMoney feat. Taelor Yung"),
		Track(
				"DNE feat. G'Natt",
				"Go Blaze",
				216704447,
				"Konrad Abramowicz",
				"Konrad OldMoney feat. Chanarah, Cidro Onetoo"
				),
		Track("ICHIBANCHI", "Dishonor", 495804559, "Konrad Abramowicz", "Konrad OldMoney feat. Brevner"),
		Track("Yamete", "Frost", 433713094, "Konrad Abramowicz", "Konrad OldMoney feat. Frawst"),
		Track(
				"UMVN feat. Imp Ra",
				"High School Bully",
				384046700,
				"Konrad Abramowicz",
				"Konrad OldMoney feat. Cidro Onetoo, Perry Porter"
				),
		Track(
				"DAPxFLEM",
				"NBOM",
				749562071,
				"Konrad Abramowicz",
				"Konrad OldMoney feat. Cidro Onetoo, Perry Porter"
				),
		Track("Code 137", "Suicide", 706423871, "Geno Lenardo", "Geno Lenardo feat. Zeale"),
		Track("HAPS", "Day of Dead", 279160746, "Konrad Abramowicz", "Konrad OldMoney feat. Taelor Yung"),
		Track("Knixit", "Bruzez", 960655358, "Konrad Abramowicz", "Konrad OldMoney feat. Johnny GR4VES"),
		Track(
				"Sugarcoob feat. ANAK KONDA",
				"Clip Boss",
				101795579,
				"Konrad Abramowicz",
				"Konrad OldMoney feat. Johnny GR4VES"
				),
		Track(
				"Triple-B feat. Gun-Fu",
				"PLUCK U",
				921768250,
				"Konrad Abramowicz",
				"Konrad OldMoney feat. Ded Stark"
				),
		Track("Pazoozu", "Hello Good Morning", 466119236, "Konrad Abramowicz", "Konrad OldMoney feat. S-God"),
		Track(
				"Bez Tatami feat. Gully Foyle",
				"Run The Block",
				39020019,
				"Konrad Abramowicz",
				"Konrad OldMoney feat. Taelor Yung"
				),
		Track(
				"Kyubik",
				"GR4VES",
				728377454,
				"Konrad Abramowicz",
				"Konrad OldMoney feat. Johnny GR4VES",
				),  # Check 3479
		Track(
				"Laputan Machine",
				"Warning Shots",
				761987513,
				"Konrad Abramowicz",
				"Konrad OldMoney feat. Perry Porter, Cidro Onetoo"
				),
		Track("Gorgon Madonna", "Metamorphosis", 274367834, "Yugen Blakrok", "Yugen Blakrok"),
		Track(
				"Yankee and the Brave",
				"No Save Point",
				595318736,
				"Michael Render",
				"Aniyah's Music",
				),  # Or is the artist Run The Jewels?
		Track(
				"TELO$",
				"Flacko Locko",
				344041802,
				"A$ap Rocky",
				"A$ap Rocky",
				),  # Also called 4Loko outside the game
		Track(
				"Cacimbo",
				"CCC",
				304807830,
				"Marek Walaszek & Dunia Hejneman",
				"Maro Music & Zuda",
				),  # TODO: not listed in the wiki; credits data has it under other songs not the station; TODO: check in game  # TODO: check 3506
		Track(
				"PeCero",
				"Nose Bleed",
				697396654,
				"Marek Walaszek & Okutama Akpan Etim",
				"Maro Music & Razzy Razo",
				),  # TODO: not listed in the wiki; credits data has it under other songs not the station; TODO: check in game
		]

#: 103.5 Radio PEBKAC
radio_pebkac: list[Track] = [
		Track("Error", "Bios", 254461336, "Jan Szarecki", "Louve"),
		Track("Sao Mai", "Drained", 450630849, "Rhys Fulber", "Rhys Fulber"),
		Track("Spoon Eater", "Subvert", 992013305, "Rhys Fulber", "Rhys Fulber"),
		Track(
				"Nablus",
				"Follow the White Crow",
				903459303,
				"Ivan Iusco; Elena Charbila",
				"Kid Moxie; Ivan Iusco; Elena Charbila"
				),
		Track("IOshrine", "Fake Spook", 854232133, "Kuba Sojka", "Poly Face"),
		Track("[flesh]reactor", "Move Dat", 472230949, "Kuba Sojka", "Poly Face"),
		Track("Bullet in the Head", "CANNIBALISMUS", 361407323, "Lutto Lento", "Lutto Lento"),
		Track("culteX", "La Canopée", 713386379, "Private Press", "Private Press"),
		Track("Skin<>Drifter", "Undertow Velocity", 104232706, "Private Press", "Private Press"),
		Track("Yards of the Moon", "II0I Break", 573266683, "Private Press", "Private Press", {784816738: RACES}),
		Track(
				"Retinal Scam",
				"Across the Floor",
				265347381,
				"Private Press",
				"Private Press",
				{252414463: JOHNNY_KERRY},
				),
		Track("Retinal Scam", "Gridflow", 789246932, "Private Press", "Private Press"),
		Track("Tar Hawk", "Vascular", 789477170, "Speed Dating", "Speed Dating feat. Horrid Charme"),
		Track("Tinnitus", "On My Way to Hell", 954776920, "Połoz", "Połoz", {371789278: "mus_sq011_totentanz_01"}),
		Track("Clockwork OS", "Stackoverflow", 28890981, "Bret Autrey", "Dos Era"),
		Track("Dukes of Azure", "Darkretro", 172754141, "Count", "Count"),
		]

#: 106.9 30 Principales
principales: list[Track] = [
		Track("Kartel Sonoro", "Bamo", 505436507, "Konrad Abramowicz", "Konrad OldMoney feat. Frawst"),
		Track(
				"Kartel Sonoro",
				"Dagga",
				507913744,
				"Konrad Abramowicz",
				"Konrad OldMoney feat. Cerbeus, Johnny GR4VES"
				),  # TODO: Check 3807
		Track("7 Facas", "Dinero", 7556210, "Konrad Abramowicz", "Konrad OldMoney feat. Cerbeus"),
		Track(
				"7 Facas",
				"Serpant",
				200129727,
				"Konrad Abramowicz",
				"Konrad OldMoney feat. Cerbeus, Johnny GR4VES"
				),
		Track(
				"Don Mara",
				"Tatted On My Face",
				626364837,
				"Konrad Abramowicz",
				"Konrad OldMoney feat. 37 Heartbreak"
				),
		Track("Big Machete", "Barrio", 973916952, "Konrad Abramowicz", "Konrad OldMoney"),
		Track("ChickyChickas", "Hood", 328370557, "Konrad Abramowicz", "Konrad OldMoney feat. Awrath"),
		Track("ChickyChickas", "Only Son", 398402575, "Konrad Abramowicz", "Konrad OldMoney feat. 37 Heartbreak"),
		Track(
				"Papito Gringo",
				"Muévelo / Cumbia",
				179243125,
				"David Perez",
				"David Rolas",
				),  # TODO: one file or two?
		Track("FKxU", "Muerto Trash", 858828322, "Konrad Abramowicz", "Konrad OldMoney feat. Blackheart NC"),
		Track("DJ CholoZ", "Westcoast Til I Die", 711322898, "Konrad Abramowicz", "Konrad OldMoney feat. Cerbeus"),
		]

#: 107.3 Morro Rock Radio
morro_rock_radio: list[Track] = [
		Track("Brutus Backlash", "Suffer Me", 360407664, "Chris Tapp", "The Cold Stares"),
		Track("XerzeX", "Heave Ho", 239971311, "Konrad Abramowicz", "Konrad OldMoney feat. Frawst"),
		Track("Beached Tarantula", "I Will Follow", 216269337, "Snot Abundance", "Snot Abundance"),
		Track("IBDY", "Who's Ready for Tomorrow", 688358243, "Rat Boy", "Rat Boy"),  # Pre 2.0 only?
		Track("IBDY", "Likewise", 92260188, "Rat Boy", "Rat Boy"),
		Track(
				"Rubicones",
				"Friday Night Fire Fight",
				534318699,
				"Jacques Barbot",
				"Aligns", {950011967: SOMEWHAT_DAMAGED_FLASHBACK}
				),
		Track("Rubicones", "Trauma", 146627573, "Jacques Barbot", "Aligns"),
		Track(
				"Blood and Ice",
				"Summer of 2069",
				395305747,
				"Metz",
				"Metz",
				),  # Also called Heaven's Gate when not in the game.
		Track(
				"Krushchev's Ghosts",
				"Testmaster / No Convenient Apocalypse",
				826126133,
				"Pissed Jeans",
				"Pissed Jeans",
				),
		Track("Cutthroat", "Sustain/Decay", 82962273, "Piotr Maciejewski", "Drivealone"),
		Track("Artificial Kids", "To the Fullest", 521529624, "The Unfit", "The Unfit"),
		Track("Fingers and the Outlaws", "So It Goes", 533813975, "Man Man", "Man Man"),
		Track(
				"SAMURAI",
				"Never Fade Away",
				230331069,
				"David Sandström; Kristofer Steen; P.T. Adamczyk; Dennis Lyxzén",
				"Refused",
				{904252670: "mus_q108_concert_glitch"},
				),
		Track(
				"SAMURAI",
				"Black Dog",
				847868585,
				"David Sandström; Kristofer Steen; Mattias Bärj; Dennis Lyxzén",
				"Refused",
				{661946813: "mus_q204_js_apartment_vinyl_01, during New Dawn Fades"},
				),
		Track(
				"SAMURAI",
				"Chippin' In",
				67866068,
				"David Sandström; Kristofer Steen; Dennis Lyxzén",
				"Refused",
				{
						1062521678: "mus_q110_chippinin",
						460511222: "mus_e3demo_end",
						350111750: "mus_q115_chippin_in",
						28560079: "mus_q001_v_apartment_chippin_in",
						961827496: "mus_q110_chippinin"
						}
				),
		Track(
				"SAMURAI",
				"The Ballad Of Buck Ravers",
				336551383,
				"David Sandström; Kristofer Steen; P.T. Adamczyk; Dennis Lyxzén",
				"Refused",
				{47509183: JOHNNY_KERRY},  # Also mus_q204_js_apartment_vinyl_01, during New Dawn Fades
				),
		]

# TODO: check against credits
#: 107.5 Dark Star
dark_star: list[Track] = [
		Track(
				"Mr. Kipper",
				"Choke Hold",
				30149432,
				'',
				"Idris Elba",
				{
						938547930: MUS_OW_PHONE,
						527795061: "mus_mq301_car_cmb_source",
						904203946: "mus_mq301_car_cmb_source"
						}
				),
		Track("BADPANNINI", "Headrush", 661309522, '', "Backxwash"),
		Track(
				"No Strings Attached",
				"Orbital Insertion",
				747444235,
				'',
				"Hagop Tchaparian", {150643485: "mus_q303_restaurant_01"}
				),
		Track("Bwana Mungu", "Cyko Arctic", 750285308, '', "Lord Spikeheart, Chrisman"),
		Track("OLO Y", "Pierwszy raz naprawdę", 22550244, '', "27.Fuckdemons", {211308495: MUS_OW_PHONE}),
		Track("DJ papergekko", "NUCLEAR DREAMLAND", 329846451, '', "julek ploski", {503455989: MUS_OW_PHONE}),
		Track("Mr. Kipper", "Rollacoaster", 569092571, '', "Idris Elba"),  # AKA Rolla
		Track(
				"Ecko Frequency",
				"IKARI",
				307219088,
				'',
				"Ecko Bazz, Silkback",
				{615284999: MUS_OW_PHONE},
				),  # AKA Zuuka
		Track("Her Mashewsky", "fabrica KOSMOS", 766230008, '', "Kuba Wandacjpwocz", {308385266: MUS_OW_PHONE}),
		Track("Walt Air", "dRk", 554215322, '', "Krzysztof Freeze Ostrowski"),
		Track("Mightonauts", "Minion Sex", 483528050, '', "T'len Lai"),
		Track("DJ papergekko", "Bigger Crimes", 71696280, '', "julek ploski"),
		Track("Łotr", "Memories of Mzuzu", 617780730, '', "Połoz"),
		]

#: Other songs.
misc: list[Track] = [
		Track(
				"P.T. Adamczyk & Sora Lion",
				"Hardest To Be",
				1032568254,
				"P.T. Adamczyk & Sora Lion",
				other_uses={
						762143559: HANGOUT_QUEST,
						44152296: "mus_ep1_credits",
						},  # 1032568254: "mus_q304_alex_heart_to_heart",
				),  # TODO: find correct main ID
		Track("P.T. Adamczyk & Dawid Podsiadło", "Phantom Liberty", 904740609, "P.T. Adamczyk, Dawid Podsiadło"),
		# TODO: Track("SAMURAI", "Archangel", 0, "David Sandström; Kristofer Steen; Dennis Lyxzén", "Refused"),
		Track("Kerry Eurodyne", "Boat Song", 341588624, '', ''),  # TODO: credits
		Track(
				"Kerry Eurodyne",
				"Chippin' In",
				750744602,
				"Borys Pugacz-Muraszkiewicz; P.T. Adamczyk",
				"Damian Ukeje; P.T. Adamczyk"
				),
		Track(
				"Footage Missing",
				"When It's War",
				925140962,
				"Trevor Samuels & Marek Walaszek",
				"Deadly Hunta & Maro Music"
				),  # TODO: check IDs; had extra ID but it was rain sounds
		Track("Erik Satie", "Gymnopédie No. 1", 776355648, "Erik Satie"),
		Track("Frédéric Chopin", "Nocturne Op. 55 No. 1", 768869823, "Frédéric Chopin"),
		Track("Claude Debussy", "Clair de Lune", 273382915, "Claude Debussy"),
		Track("P.T. Adamczyk feat. Olga Jankowska", "Never Fade Away", 1067066079, '', ''),  # TODO: credits
		Track(
				"SAMURAI",
				"A Like Supreme",
				905498026,  # mus_q204_js_apartment_vinyl_01, during New Dawn Fades
				"David Sandström; Kristofer Steen; Dennis Lyxzén",
				"Refused",
				),
		# TODO: Track("Nomad Spirit", "Katarzyna Kraińska", 0, "P.T. Adamczyk"),
		Track(
				"Baron Celine",
				"RATATATA",
				198353675,
				"Marek Aureliusz Teodoruk & Baron Black",
				"Baron Black & Auer"
				),  # mus_custom_radio_ratatata and mus_q110_voodooboys_market; plays on radio in Pacifica  # TODO: credits
		# TODO: Track("Guilt Code", "Neuron", 0, "Marek Aureliusz Teodoruk", "Auer"),
		Track(
				"CD Projekt Red",
				"Radio Synthwave GNH1",
				125084733,
				),  # TODO: find which station these play on (with copyrighted music disabled)
		Track(
				"CD Projekt Red",
				"Radio Synthwave GNH2",
				714642933,
				),  # TODO: find which station these play on (with copyrighted music disabled)
		Track(
				"CD Projekt Red",
				"Radio Synthwave GNH3",
				796147433,
				),  # TODO: find which station these play on (with copyrighted music disabled)
		Track(
				"CD Projekt Red",
				"Radio Synthwave GNH4",
				295168116,
				),  # TODO: find which station these play on (with copyrighted music disabled)
		Track(
				"CD Projekt Red",
				"Radio Synthwave GNH5",
				718643010,
				),  # TODO: find which station these play on (with copyrighted music disabled)
		Track(
				"CD Projekt Red",
				"mus_radio_03_electind_ppgame18",
				176501448,
				),  # TODO should be radio, maybe with copyrighted music disabled. Find station
		Track(
				"CD Projekt Red",
				"mus_radio_03_elec_ind_neuron",
				805938280,
				),  # TODO should be radio, maybe with copyrighted music disabled. Find station
		Track(
				"Stryjo",
				"Nieczuły parostatek",
				379273238,
				'',
				'',
				{594710394: HANGOUT_QUEST},
				),  # 1664 plays on Jazz station if copyrighted music disabled  # TODO: find main ID
		Track(
				"CD Projekt Red",
				"Laura (Instrumental)",
				1015556788,
				"David Raksin",
				),  # if copyrighted music disabled; mus_radio_08_jazz_laura  # TODO: artist
		Track(
				"CD Projekt Red",
				"You Don't Know What Love Is (Instrumental)",
				403896296,
				"Gene Depaul; Don Raye",
				other_uses={1041056699: HANGOUT_QUEST},
				),  # if copyrighted music disabled  # TODO: artist
		]

#: Mapping of station names to tracklists.
radio_stations: dict[str, list[Track]] = {
		"88.9 Pacific Dreams": pacific_dreams,
		"89.3 Radio Vexelstrom": radio_vexelstrom,
		"89.7 Growl FM": growl_fm,
		"91.9 Royal Blue Radio": royal_blue_radio,
		"92.9 Night FM": night_fm,
		"95.2 Samizdat Radio": samizdat_radio,
		"96.1 Ritual FM": ritual_fm,
		"98.7 Body Heat Radio": body_heat_radio,
		"99.9 Impulse": impulse,
		"101.9 The Dirge": the_dirge,
		"103.5 Radio PEBKAC": radio_pebkac,
		"106.9 30 Principales": principales,
		"107.3 Morro Rock Radio": morro_rock_radio,
		"107.5 Dark Star": dark_star,
		"misc": misc,
		}

radio_jingle_ids: dict[str, Sequence[int]] = {
		"101.9 The Dirge": (789210352, 53171054, 994411021),
		"107.5 Dark Star": (165565468, 771183218, 478850173, 99989307),
		"98.7 Body Heat Radio": (679337608, 936720723, 915927353),
		"95.2 Samizdat Radio": (789557574, 848658177, 957429891),
		"92.9 Night FM": (259224498, 236812550, 818146679, 837456452, 124957938),
		"96.1 Ritual FM": (752675320, 686466489, 88070376, 180892104),
		"88.9 Pacific Dreams": (128841128, 750587894, 1004904231),
		"89.3 Radio Vexelstrom": (130404871, 885647302, 339209955, 938562027),
		"106.9 30 Principales": (91143381, 922727393, 25396109),
		"91.9 Royal Blue Radio": (197336484, 66052711, 272220110, 119313970, 712496434),
		"103.5 Radio PEBKAC": (615874745, 493162163, 445903319),
		}
"""
Wem file names for radio station jingles.

Files are located in ``audio_2_soundbanks.archive`` at ``base\\sound\\soundbanks\\media\\{wem_name}.wem``
"""



#!/usr/bin/env python3
#
#  io.py
"""
File IO operations.
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import binascii
import inspect
import struct
import warnings
from collections.abc import Iterator
from typing import IO, Any, NamedTuple, TypeVar

# 3rd party
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike

# this package
from cp2077_extractor.cr2w.datatypes import Chunk, lookup_type
from cp2077_extractor.cr2w.utils import get_names_list
from cp2077_extractor.utils import decompress

# this package
from .header_structs import (
		CR2WBufferInfo,
		CR2WEmbeddedInfo,
		CR2WExportInfo,
		CR2WFile,
		CR2WFileHeader,
		CR2WFileInfo,
		CR2WImport,
		CR2WImportInfo,
		CR2WMetadata,
		CR2WNameInfo,
		CR2WProperty,
		CR2WPropertyInfo,
		CR2WTable,
		Struct
		)

__all__ = [
		"CNameError",
		"ParsingData",
		"parse_cr2w_buffer",
		"parse_cr2w_file",
		"read_buffer",
		"read_c_name",
		"read_chunk",
		"read_file_info",
		"read_struct",
		"read_tables",
		]

_S = TypeVar("_S", bound=Struct)


def read_tables(fp: IO, table_struct: type[_S], header: CR2WTable) -> Iterator[_S]:
	"""
	Read a tables of the given type in from the opened file.

	:param fp:
	:param table_struct:
	:param header:

	:returns: An iterator over instances of ``table_struct``.
	"""

	table_bytes = fp.read(table_struct._size * header.item_count)
	crc32 = binascii.crc32(table_bytes)
	assert crc32 == header.crc32, (crc32, header.crc32)
	for idx in range(header.item_count):
		chunk = table_bytes[0 + (idx * table_struct._size):table_struct._size + (idx * table_struct._size)]
		yield table_struct(*struct.unpack(table_struct._struct_format, chunk))


class CNameError(Exception):
	"""
	Error raised when an invalid name is read.
	"""


def read_c_name(fp: IO, names_list: list[bytes]) -> bytes:
	"""
	Read a name from the open file.

	Reads the ordinal of the name, and looks up the name string in ``names_list``.

	:param fp:
	:param names_list: Ordered list of names used in the file, for lookups.
	"""

	string_index = struct.unpack("<H", fp.read(2))[0]
	assert string_index < len(names_list)
	c_name = names_list[string_index]
	assert c_name
	if c_name == b"None":
		raise CNameError()
	return c_name


def read_struct(fp: IO, struct_type: type[_S]) -> _S:
	"""
	Read the given struct from the open file.

	:param fp:
	:param struct_type:
	"""

	return struct_type(*struct.unpack(struct_type._struct_format, fp.read(struct_type._size)))


def read_file_info(fp: IO) -> CR2WFileInfo:
	"""
	Read the file header and metadata.

	:param fp:
	"""

	magic = fp.read(4)
	assert magic == b"CR2W"

	# File Header
	file_header = read_struct(fp, CR2WFileHeader)  # type: ignore[type-var]

	if file_header.version > 195 or file_header.version < 163:
		raise ValueError("Unsupported Version")

	# Tables [7-9] are not used in cr2w so far.
	table_headers = [read_struct(fp, CR2WTable) for _ in range(10)]  # type: ignore[type-var]

	# Read strings - block 1 (index 0)
	assert fp.tell() == table_headers[0].offset, (fp.tell(), table_headers[0].offset)

	string_dict: dict[int, bytes] = {}
	while fp.tell() < (table_headers[0].offset + table_headers[0].item_count):
		pos = fp.tell() - table_headers[0].offset
		string = b''
		while True:
			char = fp.read(1)
			if char == b"\0":
				break
			string += (char)
		if not string:
			string = b"None"
		string_dict[pos] = string

	# Read the other tables
	name_info: list[CR2WNameInfo] = list(read_tables(fp, CR2WNameInfo, table_headers[1]))  # type: ignore[type-var]
	import_info: list[CR2WImportInfo] = list(
			read_tables(fp, CR2WImportInfo, table_headers[2])  # type: ignore[type-var]
			)
	property_info: list[CR2WPropertyInfo] = list(
			read_tables(fp, CR2WPropertyInfo, table_headers[3])  # type: ignore[type-var]
			)
	export_info: list[CR2WExportInfo] = list(
			read_tables(fp, CR2WExportInfo, table_headers[4])  # type: ignore[type-var]
			)
	buffer_info: list[CR2WBufferInfo] = list(
			read_tables(fp, CR2WBufferInfo, table_headers[5])  # type: ignore[type-var]
			)
	embedded_info: list[CR2WEmbeddedInfo] = list(
			read_tables(fp, CR2WEmbeddedInfo, table_headers[6])  # type: ignore[type-var]
			)

	_names_list: list[bytes] = []
	for a_name_info in name_info:
		assert a_name_info.offset in string_dict
		_names_list.append(string_dict[a_name_info.offset])

	_imports_list = []
	for an_import_info in import_info:
		assert an_import_info.offset in string_dict
		ret = CR2WImport(
				class_name=_names_list[an_import_info.class_name],
				depot_path=b'',  # TODO:  = depot_path or '',
				flags=an_import_info.flags,
				)
		_imports_list.append(ret)

	return CR2WFileInfo(
			file_header=file_header,
			string_dict=string_dict,
			name_info=name_info,
			import_info=import_info,
			property_info=property_info,
			export_info=export_info,
			buffer_info=buffer_info,
			embedded_info=embedded_info,
			imports=_imports_list,
			)


def read_chunk(fp: IO, chunk_index: int, file_info: CR2WFileInfo) -> tuple[bytes, bytes]:
	"""
	Read an export chunk from the file.

	:param fp:
	:param chunk_index:
	:param file_info:

	:returns: A tuple of the raw chunk data and the chunk's datatype.
	"""

	names_list = get_names_list(file_info)

	info = file_info.export_info[chunk_index]
	red_type_name = names_list[info.class_name]

	assert fp.tell() == info.data_offset
	data = fp.read(info.data_size)

	if (fp.tell() - info.data_offset != info.data_size):
		warnings.warn("Chunk size mismatch! Could lead to problems")
		fp.seek(info.data_offset + info.data_size)

	return data, red_type_name


def read_buffer(fp: IO, info: CR2WBufferInfo) -> bytes:
	"""
	Read a buffer from the CR2W/W2RC file.

	:param fp:
	:param info: Metadata about the buffer
	"""

	assert fp.tell() == info.offset

	# buffer = fp.read(info.disk_size)
	buffer = fp.read(info.mem_size)

	if buffer[:4] == b"KARK":
		# Compressed with oodle
		decompressed_size = int.from_bytes(buffer[4:8], "little")
		buffer = decompress(buffer[8:], decompressed_size)

	# TODO: check crc32 (figure out what the input data is)
	# crc32 = binascii.crc32(buffer)
	# assert crc32 == info.crc32, (crc32, info.crc32)

	return buffer


class ParsingData(NamedTuple):
	"""
	Working data for parsing CR2W/W2RC files.
	"""

	#: Name lookup table for the file.
	names_list: list[bytes]

	#: List of tuples of the raw chunk data and the chunk's datatype
	chunks: list[tuple[bytes, bytes]]

	#: List of tuples of the raw buffer data and the buffer metadata
	buffers: list[tuple[bytes, CR2WBufferInfo]]


def parse_cr2w_file(filename: PathLike) -> CR2WFile:
	"""
	Parse a CR2W/W2RC file from the given path.

	:param filename:
	"""

	filename_p = PathPlus(filename)
	with filename_p.open("rb") as fp:
		return parse_cr2w_buffer(fp, filename_p)


def parse_cr2w_buffer(fp: IO, filename: PathLike | None = None) -> CR2WFile:
	"""
	Parse a CR2W/W2RC file from an opened file.

	:param fp:
	:param filename: Optionally, the path of the opened file for inclusion in metadata.
	"""

	info = read_file_info(fp)
	assert info.string_dict, "Malformed file"

	# # TODO:
	hash_version = None
	# # use 1st string as field 0 is always empty
	# hash_version = identify_hash(info.string_dict[1], info.name_info[1].hash)
	# if (hash_version == HashVersion.Unknown):
	# 	raise ValueError("Failed to identify hash version")

	properties: list[CR2WProperty] = []
	for property_info in info.property_info:
		# TODO: properties.append(read_property(property_info))
		properties.append(CR2WProperty())

	if not properties:
		raise ValueError("Found unsupported PropertyInfo")

	# TODO: ensure CHandle/CWeakHandle can be resolved

	chunks: list[tuple[bytes, bytes]] = []

	for i in range(len(info.export_info)):
		chunks.append(read_chunk(fp, i, info))

	buffer_data: list[tuple[bytes, CR2WBufferInfo]] = []

	for buffer_info in info.buffer_info:
		buffer_data.append((read_buffer(fp, buffer_info), buffer_info))

	parsing_data = ParsingData(get_names_list(info), chunks, buffer_data)

	root_chunk_type = chunks[0][1]
	var_type = lookup_type(root_chunk_type)
	assert inspect.isclass(var_type)
	assert issubclass(var_type, Chunk)
	root_chunk = var_type.from_chunk(chunks[0][0], parsing_data)

	# TODO: read embedded files
	embedded_files: list[Any] = []  # TODO: value type
	# for embedded_info in info.embedded_info:
	# 	embedded_files.Add(read_embedded(embedded_info))

	# TODO: check fp.tell() against header field giving file length (if there is one)
	rem = fp.read(999999)
	if len(rem) != 0:
		warnings.warn(f"{len(rem)} bytes remaining in file!")

	if filename:
		meta_filename = PathPlus(filename).abspath().as_posix()
	else:
		meta_filename = None

	metadata = CR2WMetadata(
			file_name=meta_filename,
			version=info.file_header.version,
			build_version=info.file_header.build_version,
			objects_end=info.file_header.objects_end,
			hash_version=hash_version,
			)

	return CR2WFile(
			info=info,
			metadata=metadata,
			properties=properties,
			root_chunk=root_chunk,
			embedded_files=embedded_files,
			)



#!/usr/bin/env python3
#
#  physics.py
"""
Classes to represent datatypes within CR2W/W2RC files (prefixed ``physics``).
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from dataclasses import dataclass, field

# this package
from cp2077_extractor.cr2w import enums
from cp2077_extractor.cr2w.datatypes.base import Chunk, Quaternion

__all__ = [
		"physicsCustomFilterData",
		"physicsFilterData",
		"physicsMaterialReference",
		"physicsQueryFilter",
		"physicsQueryPreset",
		"physicsRagdollBodyInfo",
		"physicsRagdollBodyNames",
		"physicsSimulationFilter",
		]


@dataclass
class physicsRagdollBodyInfo(Chunk):
	parent_anim_index: int = -1
	child_anim_index: int = -1
	parent_body_index: int = -1
	body_part: enums.physicsRagdollBodyPartE = enums.physicsRagdollBodyPartE.HEAD  # TODO: CBitField
	shape_type: enums.physicsRagdollShapeType = enums.physicsRagdollShapeType.CAPSULE
	shape_radius: float = 0.05
	half_height: float = 0.05
	shape_local_translation: tuple[float, float, float] = (0.0, 0.0, 0.0)
	shape_local_rotation: Quaternion = field(default_factory=Quaternion)
	is_root_displacement_part: bool = False
	swing_angles_y: list[float] = field(default_factory=list)
	swing_angles_z: list[float] = field(default_factory=list)
	twist_angles: list[float] = field(default_factory=list)
	is_stiff: bool = False
	exclude_from_early_collision: bool = False
	filter_data_override: str = ''


@dataclass
class physicsRagdollBodyNames(Chunk):
	parent_anim_name: str = ''
	child_anim_name: str = ''


@dataclass
class physicsQueryPreset(Chunk):
	preset_name: str = ''


@dataclass
class physicsSimulationFilter(Chunk):
	mask1: int = 0
	mask2: int = 0


@dataclass
class physicsCustomFilterData(Chunk):
	collision_type: list[str] = field(default_factory=list)
	collide_with: list[str] = field(default_factory=list)
	query_detect: list[str] = field(default_factory=list)


@dataclass
class physicsMaterialReference(Chunk):
	name: str = ''


@dataclass
class physicsQueryFilter(Chunk):
	mask1: int = 0
	mask2: int = 0


@dataclass
class physicsFilterData(Chunk):
	simulation_filter: physicsSimulationFilter = field(default_factory=physicsSimulationFilter)
	query_filter: physicsQueryFilter = field(default_factory=physicsQueryFilter)
	preset: str = ''
	custom_filter_data: physicsCustomFilterData = field(default_factory=physicsCustomFilterData)



#!/usr/bin/env python3
#
#  radio_dj.py
"""
Helpers for handling radio DJ logic and audio files.
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import itertools
from collections.abc import Collection, Iterator
from typing import TYPE_CHECKING, Any, NamedTuple

# 3rd party
import networkx  # type: ignore[import-untyped]
from networkx import Graph, all_simple_paths

# this package
from cp2077_extractor.cr2w.datatypes.base import HandleData
from cp2077_extractor.cr2w.datatypes.scn import (
		scnDialogLineEvent,
		scnRewindableSectionNode,
		scnSceneGraph,
		scnSceneResource,
		scnscreenplayDialogLine,
		scnSectionNode
		)
from cp2077_extractor.cr2w.header_structs import CR2WFile

if TYPE_CHECKING:
	# 3rd party
	from matplotlib.figure import Figure  # nodep

__all__ = [
		"DJData",
		"EventData",
		"find_graph_entry_points",
		"get_link_paths",
		"load_events_dict",
		"parse_radio_scene_graph",
		"parse_subtitles",
		"plot_graph",
		"remove_intermediate_nodes",
		]


class DJData(NamedTuple):
	"""
	Data for a radio DJ's sound files.
	"""

	#: The filenames in the game files (in ``base/localization/en-us/vo`` or ``base\localization\common\vo``) for this advert all start with this string.
	audio_filename_prefix: str

	#: The name of the scene file in ``base/media/animated_billboards/scenes`` in the game files (with the suffix ``.scene``.)
	scene_file: str

	#: The name of the radio station etc.
	station_name: str

	general_audio: bool = False
	"""
	Whether the audio is common too all game languages.

	:py:obj:`False` for localised audio, :py:obj:`True` for files in ``audio_1_general.archive``.
	"""


class EventData(NamedTuple):
	"""
	Metadata for an event triggered by a node in a scene.
	"""

	event_id: str

	#: Identifier of the ``scnscreenplayDialogLine`` for this event (which indicates the file to play and the subtitle line).
	screenplay_line_id: int

	#: The suffix of the audio file, to give the file to play. e.g. `f_1BAAA2A79044D000`.
	audio_file_suffix: str

	#: Identifier of the subtitles for the event.
	subtitle_ruid: str


def load_events_dict(events_dict: dict[str, list[tuple[str, int, str, str]]]) -> dict[int, list[EventData]]:
	"""
	Load a JSON serialised mapping of node numbers to :class:`~.EventData`.

	Converts string keys back into integers and value tuples bacm into :class:`EventData`.

	:param events_dict:
	"""

	return {int(k): [EventData(*vv) for vv in v] for k, v in events_dict.items()}


def remove_intermediate_nodes(graph: Graph, audio_nodes: Collection[int]) -> None:
	"""
	Remove non-audio nodes from a networkx graph, and reconnect edges.

	:param graph:
	:param audio_nodes:
	"""

	audio_nodes_set = set(audio_nodes)
	intermediate_nodes = set(graph.nodes()) - audio_nodes_set

	for node in intermediate_nodes:

		for in_src, _ in graph.in_edges(node):
			for _, out_dst in graph.out_edges(node):
				graph.add_edge(in_src, out_dst)

		graph.remove_node(node)

	assert not set(graph.nodes()) - audio_nodes_set


def plot_graph(graph: Graph) -> "Figure":
	"""
	Plot the graph, in a hierarchy from start nodes down to finish nodes.

	Requires ``matplotlib`` and ``pygraphviz``.

	:param graph:
	"""

	# 3rd party
	from matplotlib import pyplot as plt  # nodep
	from networkx.drawing.nx_agraph import graphviz_layout  # type: ignore[import-untyped]

	pos = graphviz_layout(graph, "dot")
	fig, ax1 = plt.subplots()
	networkx.draw(graph, ax=ax1, with_labels=True, font_weight="bold", pos=pos)
	return fig


def find_graph_entry_points(graph: Graph) -> tuple[list[int], list[int], list[int]]:
	"""
	Find entry points into the radio logic graph.

	That is, any lone nodes, any start nodes, and any end nodes (skipping chatter).

	:param graph:
	"""

	lone_nodes = []
	start_nodes = []
	end_nodes = []
	for node in graph.nodes():
		if graph.in_degree(node) == 0:
			if graph.out_degree(node) == 0:
				lone_nodes.append(node)
			else:
				start_nodes.append(node)
		elif graph.out_degree(node) == 0:
			end_nodes.append(node)
		# else:
		# 	raise ValueError(node, "Intermediate node")

	return lone_nodes, start_nodes, end_nodes


def parse_subtitles(data: dict[str, Any] | CR2WFile) -> dict[str, str]:
	"""
	Parse subtitle data.

	:param data: A REDengine ``.scene`` file, either as a JSON representation as parsed by Wolvenkit, or as a :class:`~.CR2WFile` instance.
	"""

	# TODO: handle non-localised scene files with separate subtitles file.

	if isinstance(data, CR2WFile):
		return _parse_subtitles_cr2wfile(data)
	else:
		return _parse_subtitles_json(data)


def _parse_subtitles_json(scene_json: dict[str, Any]) -> dict[str, str]:
	"""
	Parse subtitle data.

	:param scene_json: JSON representation of a REDengine ``.scene`` file (as parsed by Wolvenkit).
	"""

	vp_entries = {}
	subtitles = {}
	for entry in scene_json["Data"]["RootChunk"]["locStore"]["vpEntries"]:
		vp_entries[entry["variantId"]["ruid"]] = entry["content"]
	for entry in scene_json["Data"]["RootChunk"]["locStore"]["vdEntries"]:
		subtitles[entry["locstringId"]["ruid"]] = vp_entries[entry["variantId"]["ruid"]]

	return subtitles


def _parse_subtitles_cr2wfile(cr2w_file: CR2WFile) -> dict[str, str]:
	"""
	Parse subtitle data.

	:param cr2w_file: Parsed REDengine ``.scene`` file.
	"""

	assert isinstance(cr2w_file.root_chunk, scnSceneResource)
	root_chunk: scnSceneResource = cr2w_file.root_chunk

	vp_entries = {}
	subtitles = {}
	for vp_entry in root_chunk.loc_store.vp_entries:
		vp_entries[str(vp_entry.variant_id.ruid)] = vp_entry.content
	for vd_entry in root_chunk.loc_store.vd_entries:
		subtitles[str(vd_entry.locstring_id.ruid)] = vp_entries[str(vd_entry.variant_id.ruid)]

	return subtitles


def parse_radio_scene_graph(data: dict[str, Any] | CR2WFile) -> tuple[Graph, dict[int, list[EventData]]]:
	"""
	Partial parsing of scene graph.

	Only finds dialogue events and the paths between them; no conditional logic.

	:param data: A REDengine ``.scene`` file, either as a JSON representation as parsed by Wolvenkit, or as a :class:`~.CR2WFile` instance.
	"""

	if isinstance(data, CR2WFile):
		return _parse_radio_scene_graph_cr2wfile(data)
	else:
		return _parse_radio_scene_graph_json(data)


def _parse_radio_scene_graph_json(scene_json: dict[str, Any]) -> tuple[Graph, dict[int, list[EventData]]]:
	"""
	Partial parsing of scene graph.

	Only finds dialogue events and the paths between them; no conditional logic.

	:param scene_json: JSON representation of a REDengine ``.scene`` file (as parsed by Wolvenkit).
	"""

	root_chunk = scene_json["Data"]["RootChunk"]

	screenplay_store_dict: dict[int, tuple[str, str]] = {}

	for line in root_chunk["screenplayStore"]["lines"]:
		line_data = (line["femaleLipsyncAnimationName"]["$value"], line["locstringId"]["ruid"])
		screenplay_store_dict[line["itemId"]["id"]] = line_data

	scene_graph_json = root_chunk["sceneGraph"]
	graph = networkx.DiGraph()
	audio_nodes: set[int] = set()
	audio_events: dict[int, list[EventData]] = {}

	for node in scene_graph_json["Data"]["graph"]:
		# pprint.pprint(node)
		# node_type = node["Data"]["$type"]
		# if node_type != "scnSectionNode":
		# 	print(node["Data"]["$type"])
		destinations = []
		for socket in node["Data"]["outputSockets"]:
			for destination in socket["destinations"]:
				destinations.append(destination["nodeId"]["id"])
				graph.add_edge(node["Data"]["nodeId"]["id"], destination["nodeId"]["id"])

		events = []
		for event in node["Data"].get("events", ()):
			if event["Data"]["$type"] == "scnDialogLineEvent":
				# events.append(event)
				events.append(
						EventData(
								event["Data"]["id"]["id"],
								event["Data"]["screenplayLineId"]["id"],
								*screenplay_store_dict[event["Data"]["screenplayLineId"]["id"]]
								)
						)
			# else:
			# 	print(event)
		if events:
			# print(">>>", node["Data"]["nodeId"]["id"], pprint.pformat(events ))
			node_id = node["Data"]["nodeId"]["id"]
			audio_nodes.add(node_id)
			audio_events[node_id] = events
		# else:
		# 	print(node)

	remove_intermediate_nodes(graph, audio_nodes)

	return graph, audio_events


def _parse_radio_scene_graph_cr2wfile(cr2w_file: CR2WFile) -> tuple[Graph, dict[int, list[EventData]]]:
	"""
	Partial parsing of scene graph.

	Only finds dialogue events and the paths between them; no conditional logic.

	:param cr2w_file: Parsed REDengine ``.scene`` file.
	"""

	assert isinstance(cr2w_file.root_chunk, scnSceneResource)
	root_chunk: scnSceneResource = cr2w_file.root_chunk

	screenplay_store_dict: dict[int, tuple[str, str]] = {}

	line: scnscreenplayDialogLine
	for line in root_chunk.screenplay_store.lines:
		line_data = (line.female_lipsync_animation_name.decode("UTF-8"), str(line.locstring_id.ruid))
		screenplay_store_dict[line.item_id.id] = line_data

	scene_graph: HandleData[scnSceneGraph] = root_chunk.scene_graph
	graph = networkx.DiGraph()
	audio_nodes: set[int] = set()
	audio_events: dict[int, list[EventData]] = {}

	for node in scene_graph["data"].graph:
		# pprint.pprint(node)
		destinations = []
		for socket in node["data"].output_sockets:
			for destination in socket.destinations:
				destinations.append(destination.node_id.id)
				graph.add_edge(node["data"].node_id.id, destination.node_id.id)

		if not isinstance(node["data"], (scnSectionNode, scnRewindableSectionNode)):  # TODO: other valid types
			continue

		events = []
		for event in node["data"].events:
			if isinstance(event["data"], scnDialogLineEvent):
				# events.append(event)
				events.append(
						EventData(
								str(event["data"].id.id),
								event["data"].screenplay_line_id.id,
								*screenplay_store_dict[event["data"].screenplay_line_id.id],
								)
						)
			# else:
			# 	print(event)
		if events:
			node_id = node["data"].node_id.id
			audio_nodes.add(node_id)
			audio_events[node_id] = events
		# else:
		# 	print(node)

	remove_intermediate_nodes(graph, audio_nodes)

	return graph, audio_events


def get_link_paths(graph: Graph) -> Iterator[list[int]]:
	"""
	Returns an iterator over possible paths through the various link segments.
	"""

	lone_nodes, start_nodes, end_nodes = find_graph_entry_points(graph)

	for node in lone_nodes:
		yield [node]
		# print(">>>", [node])

	for start, end in itertools.product(start_nodes, end_nodes):
		yield from all_simple_paths(graph, start, end)



#!/usr/bin/env python3
#
#  base.py
"""
Basic classes for representing datatypes within CR2W/W2RC files.
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import functools
import inspect
from collections.abc import Callable
from dataclasses import dataclass, field
from io import BytesIO
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

# 3rd party
from typing_extensions import TypedDict

# this package
from cp2077_extractor.cr2w import enums
from cp2077_extractor.cr2w.utils import get_array_variables, get_chunk_variables
from cp2077_extractor.utils import StringReader, to_snake_case

if TYPE_CHECKING:
	# this package
	from cp2077_extractor.cr2w.io import ParsingData

__all__ = [
		"Array",
		"Box",
		"CColor",
		"CMatrix",
		"Chunk",
		"DeferredBufferData",
		"EulerAngles",
		"HandleData",
		"Plane",
		"QsTransform",
		"Quaternion",
		"Sphere",
		"Transform",
		"handle",
		"instantiate_type",
		"lookup_type",
		"parse_array",
		"parse_chunk",
		"parse_cname_array",
		"parse_handle_array",
		"parse_string",
		"parse_string_array",
		"redTagList",
		"serialization_deferred_data_buffer",
		]

_red_type_lookup: dict[bytes, type | Callable[..., object]] = {}

_red_enum_list = enums.__all__[:]
_red_enum_list.remove("REDEnum")
for _class_name in _red_enum_list:
	_red_type_lookup[_class_name.encode("UTF-8")] = getattr(enums, _class_name)


class Chunk:
	"""
	Base class for chunks in CR2W/W2RC files; packed data containing variable names, types and values.
	"""

	def __init_subclass__(cls, *args, **kwargs):
		_red_type_lookup[cls.__name__.encode("UTF-8")] = cls

	@classmethod
	def from_cr2w_kwargs(cls, kwargs: dict[bytes, Any]) -> "Chunk":
		"""
		Construct from a mapping of REDengine variable names and values (as Python types).
		"""
		new_kwargs: dict[str, Any] = {
				to_snake_case(arg_name.decode("UTF-8")): arg_value
				for arg_name, arg_value in kwargs.items()
				}
		return cls(**new_kwargs)

	@classmethod
	def from_chunk(cls, chunk: bytes, parsing_data: "ParsingData") -> "Chunk":
		"""
		Parse raw bytes.

		:param chunk: The raw bytes.
		:param parsing_data:
		"""

		kwargs = parse_chunk(chunk, parsing_data)
		return cls.from_cr2w_kwargs(kwargs)


@dataclass
class Array:
	"""
	Type of an array in a CR2W/W2RC file, with the name of the inner type.
	"""

	value_red_type_name: bytes

	def __call__(self, value: bytes, parsing_data: "ParsingData") -> list:
		"""
		Convert ``value`` (representing an array) into a Python list.

		:param value:
		:param parsing_data:
		"""

		if self.value_red_type_name == b"CName":
			return parse_cname_array(value, parsing_data)

		array_value_type = lookup_type(self.value_red_type_name)
		if inspect.isclass(array_value_type) and issubclass(array_value_type, Chunk):
			return [array_value_type.from_cr2w_kwargs(av) for av in parse_array(value, parsing_data)]
		elif self.value_red_type_name == b"String":
			return parse_string_array(value)
		elif self.value_red_type_name.startswith(b"handle:"):
			return parse_handle_array(value, parsing_data)
		else:
			raise NotImplementedError(array_value_type)


def lookup_type(red_type_name: bytes) -> type | Callable[..., object]:
	"""
	Lookup a Python type from its REDengine equivalent's name.

	:param red_type_name:
	"""

	if red_type_name in _red_type_lookup:
		# print("Looked up", red_type_name, "as", _red_type_lookup[red_type_name])
		return _red_type_lookup[red_type_name]
	elif red_type_name.startswith(b"array:"):
		return Array(red_type_name.split(b":", 1)[1])
	elif red_type_name.startswith(b"handle:"):
		return handle
	else:
		raise NotImplementedError(red_type_name)


def parse_chunk(chunk: bytes, parsing_data: "ParsingData") -> dict[bytes, Any]:
	"""
	Parse the given chunk of data and return a mapping of variable names to values.

	:param chunk:
	:param parsing_data:
	"""

	variables = get_chunk_variables(chunk, parsing_data.names_list)

	kwargs: dict[bytes, Any] = {}
	for (var_c_name, red_type_name, value) in variables:
		kwargs[var_c_name] = instantiate_type(red_type_name, value, parsing_data)

	return kwargs


def parse_array(chunk: bytes, parsing_data: "ParsingData") -> list[dict[bytes, str]]:
	"""
	Parse the given chunk of data as an array and return a list of mapping of variable names to values.

	:param chunk:
	:param parsing_data:
	"""

	variables = get_array_variables(chunk, parsing_data.names_list)

	array_contents = []
	for array_item in variables:

		kwargs: dict[bytes, Any] = {}
		for (var_c_name, red_type_name, value) in array_item:
			kwargs[var_c_name] = instantiate_type(red_type_name, value, parsing_data)

		array_contents.append(kwargs)

	return array_contents


def instantiate_type(red_type_name: bytes, value: bytes, parsing_data: "ParsingData") -> object:
	"""
	Create a Python class instance for the given REDengine type and the given value.

	:param red_type_name:
	:param value:
	:param parsing_data:
	"""

	# this package
	from cp2077_extractor.cr2w.io import read_c_name

	if red_type_name == b"CName":
		return read_c_name(BytesIO(value), parsing_data.names_list)

	var_type = lookup_type(red_type_name)

	if inspect.isclass(var_type) and issubclass(var_type, enums.REDEnum):
		return var_type.from_red_name(parsing_data.names_list[uint(value)])
	elif var_type is Chunk:
		return (red_type_name, parse_chunk(value, parsing_data))
	elif isinstance(var_type, Array):
		return var_type(value, parsing_data)
	elif inspect.isclass(var_type) and issubclass(var_type, Chunk):
		return var_type.from_chunk(value, parsing_data)
	elif var_type in {handle, serialization_deferred_data_buffer}:
		return var_type(value, parsing_data)
	else:
		return var_type(value)


HandleVarType = TypeVar("HandleVarType", bound=Chunk)


class HandleData(TypedDict, Generic[HandleVarType]):
	"""
	Return type of :func:`~.handle`.
	"""

	handle_id: int
	data: HandleVarType


def handle(handle: bytes, parsing_data: "ParsingData") -> HandleData[Chunk]:
	"""
	A handle points to the data in another chunk. Read that chunk and return the resulting data.

	:param handle: Raw bytes of the handle (the value of a ``handle:xxxxxx`` type), referring to the target chunk.
	:param parsing_data:
	"""

	handle_idx = int.from_bytes(handle, "little") - 1
	chunk = parsing_data.chunks[handle_idx]
	return {"handle_id": handle_idx, "data": cast(Chunk, instantiate_type(chunk[1], chunk[0], parsing_data))}


class DeferredBufferData(TypedDict):
	"""
	Return type of :func:`~.serialization_deferred_data_buffer`.
	"""

	buffer_id: int
	flags: int
	bytes: bytes


def serialization_deferred_data_buffer(
		buffer_id: bytes,
		parsing_data: "ParsingData",
		) -> DeferredBufferData:
	"""
	A ``serializationDeferredDataBuffer`` points to a buffer in the CR2W/W2RC file, containing the actual data e.g. a texture.

	:param buffer_id: The ID of the buffer. Unknown format. Currently ignored and assumed to point to the first buffer.
	:param parsing_data:
	"""

	# TODO: Two bytes. With one buffer it's 1 0.
	assert buffer_id == b"\1\0"
	buffer_idx = 0  # TODO: proper lookup implementation
	buffer, buffer_info = parsing_data.buffers[buffer_idx]
	return {"buffer_id": buffer_idx, "flags": buffer_info.flags, "bytes": buffer}


def parse_string(data: bytes) -> str:
	"""
	Parse a bytes string (which has a VLQ i32 size prefix) to a Python string.

	:param
	"""

	return StringReader(data).parse_string()


def parse_string_array(data: bytes) -> list[str]:
	"""
	Parse an array of strings.

	:param data:
	"""

	array_size = int.from_bytes(data[:4], "little")
	string_reader = StringReader(data[4:])
	return [string_reader.parse_string() for _ in range(array_size)]


def parse_cname_array(data: bytes, parsing_data: "ParsingData") -> list[bytes]:
	"""
	Parse an array of c names.

	:param data:
	"""

	# this package
	from cp2077_extractor.cr2w.io import read_c_name

	array_size = int.from_bytes(data[:4], "little")
	buffer = BytesIO(data[4:])
	return [read_c_name(buffer, parsing_data.names_list) for _ in range(array_size)]


def parse_handle_array(data: bytes, parsing_data: "ParsingData") -> list[HandleData[Chunk]]:
	"""
	Parse an array of handles (each 4 bytes long).

	:param data:
	"""

	array_size = int.from_bytes(data[:4], "little")
	array = [handle(data[4 + (4 * idx):8 + (4 * idx)], parsing_data) for idx in range(array_size)]
	return array


@dataclass
class Quaternion(Chunk):
	i: float = 0.0
	j: float = 0.0
	k: float = 0.0
	r: float = 1.0


@dataclass
class Transform(Chunk):
	position: tuple[float, float, float, float] = (0.0, 0.0, 0.0, 0.0)
	orientation: Quaternion = field(default_factory=Quaternion)


@dataclass
class redTagList(Chunk):
	tags: list[str] = field(default_factory=list)


# def uint(value: bytes) -> int:
# 	return int.from_bytes(value, byteorder="little")

uint = functools.partial(int.from_bytes, byteorder="little")

_red_type_lookup.update({
		# b"DataBuffer": bytes,  # TODO
		b"Bool": bool,
		b"String": parse_string,
		b"Uint16": uint,
		b"Uint32": uint,
		b"Uint64": uint,
		b"Uint8": uint,
		b"CRUID": uint,
		b"TweakDBID": uint,
		b"handle": handle,
		b"raRef:animAnimSet": bytes,  # TODO
		b"serializationDeferredDataBuffer": serialization_deferred_data_buffer,
		b"NodeRef": bytes,
		b"[32]Uint8": bytes,
		})


@dataclass
class EulerAngles(Chunk):
	pitch: float = 0.0
	yaw: float = 0.0
	roll: float = 0.0


@dataclass
class QsTransform(Chunk):
	translation: tuple[float, float, float, float] = (0.0, 0.0, 0.0, 0.0)
	rotation: Quaternion = field(default_factory=Quaternion)
	scale: tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0)


@dataclass
class CColor(Chunk):
	red: int = 0
	green: int = 0
	blue: int = 0
	alpha: int = 0


@dataclass
class Sphere(Chunk):
	center_radius2: tuple[float, float, float, float] = (0.0, 0.0, 0.0, 0.0)


@dataclass
class CMatrix(Chunk):
	x: tuple[float, float, float, float] = (1.0, 0.0, 0.0, 0.0)
	y: tuple[float, float, float, float] = (0.0, 1.0, 0.0, 0.0)
	z: tuple[float, float, float, float] = (0.0, 0.0, 1.0, 0.0)
	w: tuple[float, float, float, float] = (0.0, 0.0, 0.0, 1.0)


@dataclass
class Box(Chunk):
	min: tuple[float, float, float, float] = (0.0, 0.0, 0.0, 0.0)
	max: tuple[float, float, float, float] = (0.0, 0.0, 0.0, 0.0)


@dataclass
class Plane(Chunk):
	normal_distance: tuple[float, float, float, float] = (0.0, 0.0, 0.0, 0.0)



#!/usr/bin/env python3
#
#  ent.py
"""
Classes to represent datatypes within CR2W/W2RC files (prefixed ``ent``).
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import sys
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

# this package
from cp2077_extractor.cr2w import enums
from cp2077_extractor.cr2w.datatypes.appearance import appearanceAppearanceResource
from cp2077_extractor.cr2w.datatypes.base import CColor, Chunk, Plane, Quaternion, Transform, redTagList
from cp2077_extractor.cr2w.datatypes.mesh import CMesh
from cp2077_extractor.cr2w.datatypes.rend import rendSLightFlickering

if TYPE_CHECKING:
	# this package
	from cp2077_extractor.cr2w.datatypes.world import WorldTransform, worldEffect

__all__ = [
		"CIESDataResource",
		"entBaseCameraComponent",
		"entEntity",
		"entEntityID",
		"entEntityParametersBuffer",
		"entEntityTemplate",
		"entFallbackSlot",
		"entGameEntity",
		"entIAttachment",
		"entIBinding",
		"entIComponent",
		"entIPlacedComponent",
		"entIPositionProvider",
		"entISourceBinding",
		"entITransformBinding",
		"entIVisualComponent",
		"entLightComponent",
		"entSlot",
		"entSlotComponent",
		"entTagMask",
		"entTemplateAppearance",
		"entTemplateBindingOverride",
		"entTemplateComponentBackendDataOverrideInfo",
		"entTemplateComponentResolveSettings",
		"entTemplateInclude",
		"entVisualTagsSchema",
		"entVoicesetInputToBlock",
		"entdismembermentAppearanceMatch",
		"entdismembermentCullObject",
		"entdismembermentDangleInfo",
		"entdismembermentEffectResource",
		"entdismembermentFillMeshInfo",
		"entdismembermentMeshInfo",
		"entdismembermentPhysicsInfo",
		"entdismembermentWoundConfig",
		"entdismembermentWoundConfigContainer",
		"entdismembermentWoundDecal",
		"entdismembermentWoundMeshes",
		"entdismembermentWoundResource",
		"entdismembermentWoundsConfigSet",
		]


@dataclass
class entVisualTagsSchema(Chunk):
	visual_tags: redTagList = field(default_factory=redTagList)
	schema: str = ''


@dataclass
class entTemplateComponentResolveSettings(Chunk):
	component_name: str = ''
	name_param: str = ''
	mode: enums.entTemplateComponentResolveMode = enums.entTemplateComponentResolveMode.AutoSelect


@dataclass
class entTagMask(Chunk):
	hard_tags: redTagList = field(default_factory=redTagList)
	soft_tags: redTagList = field(default_factory=redTagList)
	excluded_tags: redTagList = field(default_factory=redTagList)


@dataclass
class entdismembermentCullObject(Chunk):
	plane: Plane = field(default_factory=lambda: Plane(normal_distance=(0.0, 0.0, 1.0, -0.0)))
	plane1: Plane = field(default_factory=lambda: Plane(normal_distance=(0.0, 0.0, 1.0, -0.0)))
	capsule_point_a: tuple[float, float, float] = (0.0, 0.0, 0.0)
	capsule_point_b: tuple[float, float, float] = (0.0, 0.0, 0.0)
	capsule_radius: float = 0.1
	nearest_anim_bone_name: str = ''
	nearest_anim_index: int = -1
	ragdoll_body_index: int = sys.maxsize


@dataclass
class entdismembermentAppearanceMatch(Chunk):
	character: str = ''
	mesh: str = ''
	set_by_user: bool = False


@dataclass
class entdismembermentPhysicsInfo(Chunk):
	density_scale: float = 1.0


@dataclass
class entdismembermentMeshInfo(Chunk):
	mesh: CMesh = field(default_factory=CMesh)  # TODO: CResourceAsyncReference
	mesh_appearance: str = "default"
	appearance_map: list[entdismembermentAppearanceMatch] = field(default_factory=list)
	should_receive_decal: bool = False
	body_part_mask: enums.physicsRagdollBodyPartE = enums.physicsRagdollBodyPartE.HEAD  # TODO: CBitField
	wound_type: enums.entdismembermentWoundTypeE = enums.entdismembermentWoundTypeE.CLEAN | enums.entdismembermentWoundTypeE.COARSE  # TODO: CBitField
	cull_mesh: enums.entdismembermentWoundTypeE = enums.entdismembermentWoundTypeE.CLEAN  # TODO: CBitField
	offset: Transform = field(default_factory=Transform)
	scale: tuple[float, float, float] = (1.0, 1.0, 1.0)
	physics: entdismembermentPhysicsInfo = field(default_factory=entdismembermentPhysicsInfo)


@dataclass
class entdismembermentDangleInfo(Chunk):
	dangle_segment_lenght: float = 0.1
	dangle_velocity_damping: float = 0.4
	dangle_bend_stiffness: float = 0.6
	dangle_segment_stiffness: float = 0.0
	dangle_collision_sphere_radius: float = 0.25


@dataclass
class entdismembermentFillMeshInfo(entdismembermentMeshInfo):
	placement: enums.entdismembermentPlacementE = enums.entdismembermentPlacementE.MAIN_MESH  # TODO: CBitField
	simulation: enums.entdismembermentSimulationTypeE = enums.entdismembermentSimulationTypeE.NONE
	dangle: entdismembermentDangleInfo = field(default_factory=entdismembermentDangleInfo)


@dataclass
class entdismembermentWoundMeshes(Chunk):
	resource_set: enums.entdismembermentResourceSetE = enums.entdismembermentResourceSetE.NONE
	meshes: list[entdismembermentMeshInfo] = field(default_factory=list)
	fill_meshes: list[entdismembermentFillMeshInfo] = field(default_factory=list)


@dataclass
class entdismembermentWoundDecal(Chunk):
	offset_a: tuple[float, float, float] = (0.0, 0.0, 0.0)
	offset_b: tuple[float, float, float] = (0.0, 0.0, 0.0)
	scale: float = 1.0
	fade_origin: float = 0.7
	fade_power: float = 1.0
	resource_sets: enums.entdismembermentResourceSetMask = enums.entdismembermentResourceSetMask.BARE  # TODO: CBitField
	material: Any = None  # TODO: CResourceAsyncReference<IMaterial> = field(default_factory=CResourceAsyncReference<IMaterial>)


@dataclass
class entdismembermentWoundResource(Chunk):
	name: str = ''
	wound_type: enums.entdismembermentWoundTypeE = enums.entdismembermentWoundTypeE.CLEAN | enums.entdismembermentWoundTypeE.COARSE  # TODO: CBitField
	body_part: enums.physicsRagdollBodyPartE = enums.physicsRagdollBodyPartE.HEAD  # TODO: CBitField
	cull_object: entdismembermentCullObject = field(default_factory=entdismembermentCullObject)
	garment_morph_strength: float = 1.0
	use_procedural_cut: bool = False
	use_single_mesh_for_ragdoll: bool = False
	is_critical: bool = False
	resources: list[entdismembermentWoundMeshes] = field(default_factory=list)
	decals: list[entdismembermentWoundDecal] = field(default_factory=list)
	censored_paths: list[int] = field(default_factory=list)
	censored_cooked_paths: list[Chunk] = field(default_factory=list)  # TODO: CResourceAsyncReference
	censorship_valid: bool = False


@dataclass
class entdismembermentEffectResource(Chunk):
	name: str = ''
	appearance_names: list[str] = field(default_factory=list)
	body_part_mask: enums.physicsRagdollBodyPartE = enums.physicsRagdollBodyPartE.HEAD  # TODO: CBitField
	offset: Transform = field(default_factory=Transform)
	placement: enums.entdismembermentPlacementE = enums.entdismembermentPlacementE.MAIN_MESH  # TODO: CBitField
	resource_sets: enums.entdismembermentResourceSetMask = enums.entdismembermentResourceSetMask.BARE | enums.entdismembermentResourceSetMask.BARE1 | enums.entdismembermentResourceSetMask.BARE2 | enums.entdismembermentResourceSetMask.BARE3 | enums.entdismembermentResourceSetMask.GARMENT | enums.entdismembermentResourceSetMask.GARMENT1 | enums.entdismembermentResourceSetMask.GARMENT2 | enums.entdismembermentResourceSetMask.GARMENT3  # TODO: CBitField
	wound_type: enums.entdismembermentWoundTypeE = enums.entdismembermentWoundTypeE.COARSE  # TODO: CBitField
	effect: "worldEffect" = field(
			default_factory=lambda: worldEffect()
			)  # TODO: resolve default class (circular import)    # TODO: CResourceAsyncReference
	match_to_wound_by_name: bool = False


@dataclass
class entdismembermentWoundConfig(Chunk):
	wound_name: str = ''
	resource_set: enums.entdismembermentResourceSetE = enums.entdismembermentResourceSetE.BARE


@dataclass
class entdismembermentWoundConfigContainer(Chunk):
	appearance_name: str = ''
	wounds: list[entdismembermentWoundConfig] = field(default_factory=list)


@dataclass
class entdismembermentWoundsConfigSet(Chunk):
	configs: list[entdismembermentWoundConfigContainer] = field(default_factory=list)


@dataclass
class entIBinding(Chunk):
	enabled: bool = False
	enable_mask: entTagMask = field(default_factory=entTagMask)
	bind_name: str = ''


@dataclass
class entTemplateBindingOverride(Chunk):
	component_name: str = ''
	property_name: str = ''
	binding: entIBinding = field(default_factory=entIBinding)


@dataclass
class entTemplateComponentBackendDataOverrideInfo(Chunk):
	component_name: str
	offset: tuple[int, int]


@dataclass
class entEntityParametersBuffer(Chunk):
	parameter_buffers: list[Any] = field(default_factory=list)  # TODO: SerializationDeferredDataBuffer


@dataclass
class entEntityTemplate(Chunk):
	cooking_platform: enums.ECookingPlatform = enums.ECookingPlatform.PLATFORM_None
	includes: list["entTemplateInclude"] = field(default_factory=list)
	appearances: list["entTemplateAppearance"] = field(default_factory=list)
	default_appearance: str = "default"
	visual_tags_schema: entVisualTagsSchema = field(default_factory=entVisualTagsSchema)
	component_resolve_settings: list[entTemplateComponentResolveSettings] = field(default_factory=list)
	binding_overrides: list[entTemplateBindingOverride] = field(default_factory=list)
	backend_data_overrides: list[entTemplateComponentBackendDataOverrideInfo] = field(default_factory=list)
	local_data: Any = None  # TODO: DataBuffer = field(default_factory=DataBuffer)
	include_instance_buffer: Any = None  # TODO: DataBuffer = field(default_factory=DataBuffer)
	compiled_data: Any = None  # TODO: DataBuffer = field(default_factory=DataBuffer)
	resolved_dependencies: list[Chunk] = field(default_factory=list)  # TODO: CResourceAsyncReference
	inplace_resources: list[Chunk] = field(default_factory=list)  # TODO: CResourceReference
	compiled_entity_lodflags: int = 0


@dataclass
class entTemplateAppearance(Chunk):
	name: str = ''
	appearance_resource: appearanceAppearanceResource = field(
			default_factory=appearanceAppearanceResource
			)  # TODO: CResourceAsyncReference
	appearance_name: str = ''


@dataclass
class entTemplateInclude(Chunk):
	name: str = ''
	template: entEntityTemplate = field(default_factory=entEntityTemplate)  # TODO: CResourceAsyncReference


@dataclass
class entEntity(Chunk):
	custom_camera_target: enums.ECustomCameraTarget = enums.ECustomCameraTarget.ECCTV_All
	render_scene_layer_mask: enums.RenderSceneLayerMask = enums.RenderSceneLayerMask.Default  # CBitField


@dataclass
class entGameEntity(entEntity):
	pass


@dataclass
class entIComponent(Chunk):
	name: str = ''
	is_replicable: bool = False
	id: int = 0


@dataclass
class entISourceBinding(entIBinding):
	pass


@dataclass
class entITransformBinding(entISourceBinding):
	pass


@dataclass
class entIPlacedComponent(entIComponent):
	local_transform: "WorldTransform" = field(
			default_factory=lambda: WorldTransform()
			)  # TODO: resolve default class (circular import)
	parent_transform: entITransformBinding = field(default_factory=entITransformBinding)


@dataclass
class entSlot(Chunk):
	slot_name: str = ''
	relative_position: tuple[float, float, float] = (0.0, 0.0, 0.0)
	relative_rotation: Quaternion = field(default_factory=Quaternion)
	bone_name: str = ''


@dataclass
class entFallbackSlot(Chunk):
	slot_name: str = ''
	bone_name: str = ''


@dataclass
class entSlotComponent(entIPlacedComponent):
	# TODO: name = "Component";
	# TODO: local_transform = new WorldTransform { Position = new WorldPosition { X = new FixedPoint(), Y = new FixedPoint(), Z = new FixedPoint() }, Orientation = new Quaternion { R = 1.000000F } };
	slots: list[entSlot] = field(default_factory=list)
	fallback_slots: list[entFallbackSlot] = field(default_factory=list)


@dataclass
class entVoicesetInputToBlock(Chunk):
	input: str = ''
	block_specific_variation: bool = False
	variation_number: int = 0


@dataclass
class entIPositionProvider(Chunk):
	pass


@dataclass
class entIVisualComponent(entIPlacedComponent):
	auto_hide_distance: float = 0.0
	render_scene_layer_mask: enums.RenderSceneLayerMask = enums.RenderSceneLayerMask.Default  # TODO: CBitField
	force_lodlevel: int = 0


@dataclass
class CIESDataResource(Chunk):
	cooking_platform: enums.ECookingPlatform = enums.ECookingPlatform.PLATFORM_None
	samples: list[int] = field(default_factory=list)  # new(128)


@dataclass
class entLightComponent(entIVisualComponent):
	# TODO: name = "Component";
	# TODO: local_transform = new WorldTransform { Position = new WorldPosition { X = new FixedPoint(), Y = new FixedPoint(), Z = new FixedPoint() }, Orientation = new Quaternion { R = 1.000000F } };
	# TODO: auto_hide_distance = 15.00
	# TODO: render_scene_layer_mask = Enums.RenderSceneLayerMask.Default;
	# TODO: force_lod_level = -1;
	# TODO: light_channel = Enums.rendLightChannel.LC_Channel1 | Enums.rendLightChannel.LC_Channel2 | Enums.rendLightChannel.LC_Channel3 | Enums.rendLightChannel.LC_Channel4 | Enums.rendLightChannel.LC_Channel5 | Enums.rendLightChannel.LC_Channel6 | Enums.rendLightChannel.LC_Channel7 | Enums.rendLightChannel.LC_Channel8 | Enums.rendLightChannel.LC_ChannelWorld;
	type: enums.ELightType = enums.ELightType.LT_Point
	color: CColor = field(default_factory=CColor)
	radius: float = 5.0
	unit: enums.ELightUnit = enums.ELightUnit.LU_Lumen
	intensity: float = 100.0
	ev: float = 0.0
	temperature: float = -1.0
	light_channel: enums.rendLightChannel = enums.rendLightChannel.LC_Channel1  # TODO: CBitField
	scene_diffuse: bool = True
	scene_specular_scale: int = 100
	directional: bool = False
	roughness_bias: int = 0
	scale_gi: int = 100
	scale_env_probes: int = 100
	use_in_transparents: bool = True
	scale_vol_fog: int = 0
	use_in_particles: bool = True
	attenuation: enums.rendLightAttenuation = enums.rendLightAttenuation.LA_InverseSquare
	clamp_attenuation: bool = False
	group: enums.rendLightGroup = enums.rendLightGroup.LG_Group0
	area_shape: enums.EAreaLightShape = enums.EAreaLightShape.ALS_Capsule
	area_two_sided: bool = True
	spot_capsule: bool = False
	source_radius: float = 0.05
	capsule_length: float = 1.0
	area_rect_side_a: float = 1.0
	area_rect_side_b: float = 1.0
	inner_angle: float = 30.0
	outer_angle: float = 45.0
	softness: float = 2.0
	enable_local_shadows: bool = False
	enable_local_shadows_force_statics_only: bool = False
	contact_shadows: enums.rendContactShadowReciever = enums.rendContactShadowReciever.CSR_None
	shadow_angle: float = -10.0
	shadow_radius: float = -10.0
	shadow_fade_distance: float = 10.0
	shadow_fade_range: float = 5.0
	shadow_softness_mode: enums.ELightShadowSoftnessMode = enums.ELightShadowSoftnessMode.LSSM_Default
	ray_traced_shadows_platform: enums.rendRayTracedShadowsPlatform = enums.rendRayTracedShadowsPlatform.RLSP_All
	ray_tracing_light_source_radius: float = -1.0
	ray_tracing_contact_shadow_range: float = -1.0
	ies_profile: CIESDataResource = field(default_factory=CIESDataResource)  # TODO: CResourceAsyncReference
	flicker: rendSLightFlickering = field(default_factory=rendSLightFlickering)
	env_color_group: enums.EEnvColorGroup = enums.EEnvColorGroup.ECG_Default
	color_group_saturation: int = 100
	portal_angle_cutoff: int = 0
	allow_distant_light: bool = True
	ray_tracing_intensity_scale: float = 1.0
	path_tracing_light_usage: enums.rendEPathTracingLightUsage = enums.rendEPathTracingLightUsage.PTLU_Everywhere
	path_tracing_override_scale_gi: bool = True
	rtxdi_shadow_starting_distance: float = -1.0
	is_enabled: bool = True


@dataclass
class entEntityID(Chunk):
	hash: int = 0


@dataclass
class entIAttachment(Chunk):
	source: entIComponent = field(default_factory=entIComponent)
	destination: entIComponent = field(default_factory=entIComponent)


@dataclass
class entBaseCameraComponent(entIPlacedComponent):
	# TODO: name = "Component";
	# TODO: local_transform = new WorldTransform { Position = new WorldPosition { X = new FixedPoint(), Y = new FixedPoint(), Z = new FixedPoint() }, Orientation = new Quaternion { R = 1.000000F } };
	fov: float = 60.0
	zoom: float = 1.0
	near_plane_override: float = 0.0
	far_plane_override: float = 0.0
	motion_blur_scale: float = 1.0



#!/usr/bin/env python3
#
#  redarchive_reader.py
"""
Partial parser for REDEngine ``.archive`` files.
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import hashlib
import struct
from dataclasses import dataclass
from pathlib import PureWindowsPath
from typing import IO

# 3rd party
from domdf_python_tools.typing import PathLike
from fnvhash import fnv1a_64  # type: ignore[import-untyped]

# this package
from cp2077_extractor.utils import decompress

__all__ = ["FileList", "FileRecord", "FileSegment", "REDArchive"]


@dataclass
class FileRecord:
	"""
	Represents a FileRecord in a ``.archive`` file.
	"""

	#: FNV1A64 hash of the filename
	name_hash: int

	#: Windows filetime of file creation
	timestamp: int

	#: Number of inline buffers
	num_inline_buffer_segs: int

	#: Index of the first file segment
	segs_start: int

	#: Index of the last file segment
	segs_end: int

	#: Index of the first resource dependency
	res_deps_start: int

	#: Index of the last resource dependency
	res_deps_end: int

	#: SHA1 hash of the file  # TODO: compressed files?
	sha1_hash: bytes


@dataclass
class FileSegment:
	"""
	Represents a FileSegment in a ``.archive`` file.
	"""

	#: Offset of the data
	offset: int

	#: Size of compressed data
	zsize: int

	#: Size of uncompressed data
	size: int


@dataclass
class FileList:
	"""
	Represents a FileList in a ``.archive`` file.
	"""

	#: Always 8
	file_table_offset: int

	file_table_size: int

	#: Checksum of ???
	crc: int

	#: Number of files
	file_entry_count: int

	#: Number of file segments
	file_segment_count: int

	#: Number of resource dependencies
	resource_dep_count: int

	file_records: list[FileRecord]
	file_segments: list[FileSegment]
	resource_dependencies: list[int]

	def find_filename(self, filename: str) -> FileRecord:
		"""
		Find the record for the given filename, relative to the root of the archive (usually starting ``base``).

		:param filename:
		"""

		# TODO: cache hashes and mapping of hash to records for speed
		name_hash = fnv1a_64(bytes(PureWindowsPath(filename)))
		for record in self.file_records:
			if record.name_hash == name_hash:
				return record

		raise FileNotFoundError(filename)

	def get_segments(self, file: FileRecord) -> list[FileSegment]:
		"""
		Returns the segments for the given file.

		:param file:
		"""

		return self.file_segments[file.segs_start:file.segs_end]


@dataclass
class REDArchive:
	"""
	Represents a REDEngine ``.archive`` file.
	"""

	#: Constant: "RDAR"
	magic: str

	#: Currently 12
	version: int

	#: Offset of beginning of file list
	index_pos: int

	#: Size of file list
	index_size: int

	#: Always 0
	debug_pos: int

	#: Always 0
	debug_size: int

	#: Size of file (excluding Filesize)
	filesize: int

	#: Files created with WolvenKit only
	custom_data_length: int

	file_list: FileList

	@classmethod
	def load_archive(cls, archive_file: PathLike) -> "REDArchive":
		"""
		Load metadata for an ``.archive`` file.

		:param archive_file:
		"""

		with open(archive_file, "rb") as fp:
			magic: bytes
			magic, version, index_pos, index_size, debug_pos, debug_size, filesize, custom_data_length = struct.unpack("<4sIQIQIQI", fp.read(44))

			fp.seek(index_pos)

			file_table_offset, file_table_size, crc, file_entry_count, file_segment_count, resource_dep_count = struct.unpack("<IIQIII", fp.read(28))

			# f.read(file_table_offset)
			file_records = []
			file_segments = []

			for _ in range(file_entry_count):

				name_hash, timestamp, num_inline_buffer_segs, segs_start, segs_end, res_deps_start, res_deps_end = struct.unpack("<QqIIIII", fp.read(36))
				sha1_hash = fp.read(20)

				file_records.append(
						FileRecord(
								name_hash=name_hash,
								timestamp=timestamp,
								num_inline_buffer_segs=num_inline_buffer_segs,
								segs_start=segs_start,
								segs_end=segs_end,
								res_deps_start=res_deps_start,
								res_deps_end=res_deps_end,
								sha1_hash=sha1_hash,
								)
						)

			for _ in range(file_segment_count):
				offset, zsize, size = struct.unpack("<QII", fp.read(16))
				file_segments.append(FileSegment(offset=offset, zsize=zsize, size=size))

			resource_dependencies = list(struct.unpack(f"<{resource_dep_count}Q", fp.read(8 * resource_dep_count)))

			# remainder = fp.read(9999)
			# print(len(remainder))

			file_list = FileList(
					file_table_offset=file_table_offset,
					file_table_size=file_table_size,
					crc=crc,
					file_entry_count=file_entry_count,
					file_segment_count=file_segment_count,
					resource_dep_count=resource_dep_count,
					file_records=file_records,
					file_segments=file_segments,
					resource_dependencies=resource_dependencies,
					)

			return cls(
					magic=magic.decode("UTF-8"),
					version=version,
					index_pos=index_pos,
					index_size=index_size,
					debug_pos=debug_pos,
					debug_size=debug_size,
					filesize=filesize,
					custom_data_length=custom_data_length,
					file_list=file_list,
					)

	def extract_file(self, fp: IO, file: FileRecord) -> bytes:
		"""
		Extract a file from the archive.

		:param fp: File handle for the opened archive.
		:param file: The file to extract.
		"""

		segments = self.file_list.get_segments(file)
		compressed = False

		file_content = b''
		for segment in segments:
			fp.seek(segment.offset, 0)
			signature = fp.read(4)
			if signature == b"KARK":
				compressed = True
				# Compressed with kraken
				size = struct.unpack("<i", fp.read(4))[0]
				assert segment.size == size

				file_content += decompress(fp.read(segment.zsize - 8), size)

			else:
				file_content += signature
				assert segment.size == segment.zsize
				file_content += fp.read(segment.zsize - 4)

		if not compressed:
			# TODO: is it the sha1 of the compressed data?
			sha1_hash = hashlib.sha1(file_content).digest()
			assert sha1_hash == file.sha1_hash, (sha1_hash, file.sha1_hash)

		return file_content



#!/usr/bin/env python3
#
#  quest.py
"""
Classes to represent datatypes within CR2W/W2RC files (prefixed ``quest``).
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from dataclasses import dataclass, field

# this package
from cp2077_extractor.cr2w import enums
from cp2077_extractor.cr2w.datatypes.base import Chunk
from cp2077_extractor.cr2w.datatypes.game import gameTier3CameraSettings
from cp2077_extractor.cr2w.datatypes.graph import graphGraphNodeDefinition, graphGraphSocketDefinition
from cp2077_extractor.cr2w.datatypes.work import workWorkEntryId

__all__ = [
		"graphIGraphNodeCondition",
		"questAICommandParams",
		"questDisableableNodeDefinition",
		"questFactsDBManagerNodeDefinition",
		"questIBaseCondition",
		"questIConditionType",
		"questIFactsDBManagerNodeType",
		"questISceneConditionType",
		"questITimeConditionType",
		"questNodeDefinition",
		"questPauseConditionNodeDefinition",
		"questRealtimeDelay_ConditionType",
		"questSetVar_NodeType",
		"questSignalStoppingNodeDefinition",
		"questSocketDefinition",
		"questTimeCondition",
		"questTypedCondition",
		"questUseWorkspotParamsV1",
		"questUseWorkspotPlayerParams",
		"scnAICommandFactory",
		]


@dataclass
class scnAICommandFactory(Chunk):
	pass


@dataclass
class questAICommandParams(scnAICommandFactory):
	pass


class graphIGraphNodeCondition(Chunk):
	pass


class questIBaseCondition(graphIGraphNodeCondition):
	pass


@dataclass
class questUseWorkspotPlayerParams(Chunk):
	tier: enums.questUseWorkspotTier = enums.questUseWorkspotTier.Tier3
	camera_settings: gameTier3CameraSettings = field(default_factory=gameTier3CameraSettings)
	empty_hands: bool = False
	camera_use_trajectory_space: bool = True
	apply_camera_params: bool = False
	vehicle_procedural_camera_weight: float = 1.0
	parallax_weight: float = 1.0
	parallax_space: enums.questCameraParallaxSpace = enums.questCameraParallaxSpace.Trajectory


@dataclass
class questUseWorkspotParamsV1(questAICommandParams):
	function: enums.questUseWorkspotNodeFunctions = enums.questUseWorkspotNodeFunctions.UseWorkspot
	workspot_node: str = ''
	teleport: bool = True
	finish_animation: bool = True
	force_entry_anim_name: str = ''
	jump_to_entry: bool = False
	entry_id: workWorkEntryId = field(default_factory=workWorkEntryId)
	entry_tag: str = ''
	change_workspot: bool = True
	enable_idle_mode: bool = False
	exit_entry_id: workWorkEntryId = field(default_factory=workWorkEntryId)
	exit_anim_name: str = ''
	instant: bool = False
	is_workspot_infinite: bool = True
	is_player: bool = False
	player_params: questUseWorkspotPlayerParams = field(
			default_factory=questUseWorkspotPlayerParams
			)  # TODO: new questUseWorkspotPlayerParams { CameraSettings = new gameTier3CameraSettings { YawLeftLimit = 60.000000F, YawRightLimit = 60.000000F, PitchTopLimit = 60.000000F, PitchBottomLimit = 45.000000F, PitchSpeedMultiplier = 1.000000F, YawSpeedMultiplier = 1.000000F }, CameraUseTrajectorySpace = true, VehicleProceduralCameraWeight = 1.000000F, ParallaxWeight = 1.000000F };
	repeat_command_on_interrupt: bool = True
	work_excluded_gestures: list[workWorkEntryId] = field(default_factory=list)
	movement_type: enums.moveMovementType = enums.moveMovementType.Walk
	continue_in_combat: bool = False
	max_anim_time_limit: float = 0.0
	mesh_dissolving_enabled: bool = True
	dangle_reset_simulation: bool = False


@dataclass
class questNodeDefinition(graphGraphNodeDefinition):
	id: int = 0


class questISceneConditionType(Chunk):
	pass


@dataclass
class questDisableableNodeDefinition(questNodeDefinition):
	pass


@dataclass
class questIFactsDBManagerNodeType(Chunk):
	pass


@dataclass
class questFactsDBManagerNodeDefinition(questDisableableNodeDefinition):
	# TODO: Id = sys.maxsize
	type: questIFactsDBManagerNodeType = field(default_factory=questIFactsDBManagerNodeType)


@dataclass
class questSocketDefinition(graphGraphSocketDefinition):
	type: enums.questSocketType = enums.questSocketType.Undefined


@dataclass
class questSetVar_NodeType(questIFactsDBManagerNodeType):
	fact_name: str = ''
	value: int = 1
	set_exact_value: bool = False


@dataclass
class questSignalStoppingNodeDefinition(questDisableableNodeDefinition):
	pass


@dataclass
class questPauseConditionNodeDefinition(questSignalStoppingNodeDefinition):
	# TODO: id = sys.maxsize
	condition: questIBaseCondition = field(default_factory=questIBaseCondition)


@dataclass
class questTypedCondition(questIBaseCondition):
	pass


class questIConditionType(Chunk):
	pass


@dataclass
class questITimeConditionType(questIConditionType):
	pass


@dataclass
class questRealtimeDelay_ConditionType(questITimeConditionType):
	hours: int = 0
	minutes: int = 0
	seconds: int = 0
	miliseconds: int = 0


@dataclass
class questTimeCondition(questTypedCondition):
	type: questITimeConditionType = field(default_factory=questITimeConditionType)



#!/usr/bin/env python3
#
#  __init__.py
r"""
Data for radio stations, ambient music, etc.

Audio files are located in ``audio_2_soundbanks.archive`` at ``base\sound\soundbanks\media\{wem_name}.wem``
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from collections.abc import Iterator
from typing import Any, Protocol, SupportsIndex, overload

# this package
from .radio_stations import (
		CLOUDS_MUSIC,
		DARK_MATTER,
		HANGOUT_QUEST,
		JOHNNY_KERRY,
		MUS_OW_PHONE,
		RACES,
		SOMEWHAT_DAMAGED_FLASHBACK,
		SOMEWHAT_DAMAGED_FLASHBACK_REED,
		radio_stations
		)

__all__ = [
		"SceneAudioData",
		"intro_ids",
		"nomad_bard_ids",
		"clouds_music_ids",
		"somewhat_damaged_flashback_ids",
		"sex_cutscene_music_ids",
		"hangout_music_ids",
		"race_music_ids",
		"mus_ow_phone_ids",
		"dark_matter_music_ids",
		"johnny_kerry_music_ids",
		]

#: Mapping of wem file names to languages for the game intro audio.
intro_ids: dict[int, str] = {
		# Intros in foreign languages
		193250534: "Korean",
		212033934: "German",
		228132900: "Japanese",
		521504529: "Russian",
		533464258: "English",
		634636526: "Spanish",
		731319179: "Chinese",
		835783005: "Polish",
		964673994: "Portugese",
		1038719001: "French",
		1052911782: "Italian",
		}

# mus_q114_nomad_bard_01_song_01, plays during We Gotta Live Together (Aldecaldos ending)
#: Mapping of wem file names to language (or the guitar part) for Jake Scooter playing during the quest *We Gotta Live Together*.
nomad_bard_ids: dict[int, str] = {
		422912739: "German",  # mus_q114_nomad_bard_01_song_01/mus_q114_nomad_bard_01_song
		717901062: "Korean",  # mus_q114_nomad_bard_01_song_01/mus_q114_nomad_bard_01_song
		788187980: "English",  # mus_q114_nomad_bard_01_song_01/mus_q114_nomad_bard_01_song
		791357735: "French",  # mus_q114_nomad_bard_01_song_01/mus_q114_nomad_bard_01_song
		804002160: "Spanish",  # mus_q114_nomad_bard_01_song_01/mus_q114_nomad_bard_01_song
		845132118: "Portugese",  # mus_q114_nomad_bard_01_song_01/mus_q114_nomad_bard_01_song
		85588625: "Chinese",  # mus_q114_nomad_bard_01_song_01/mus_q114_nomad_bard_01_song
		884356362: "Japanese",  # mus_q114_nomad_bard_01_song_01/mus_q114_nomad_bard_01_song
		992577309: "Polish",  # mus_q114_nomad_bard_01_song_01/mus_q114_nomad_bard_01_song
		113512491: "Italian",  # mus_q114_nomad_bard_01_song_01/mus_q114_nomad_bard_01_song
		223185989: "Russian",  # mus_q114_nomad_bard_01_song_01/mus_q114_nomad_bard_01_song
		274240532: "Korean (2)",  # mus_q114_nomad_bard_01_song_01/mus_q114_nomad_bard_01_song
		377542210: "guitar",  # mus_q114_nomad_bard_01_song_01/mus_q114_nomad_bard_01_song
		308885281: "guitar (2)",  # mus_q114_nomad_bard_01_song
		352828163: "guitar (3)",  # mus_q114_nomad_bard_01_song
		}

# mus_q105_dollhouse
#: Music that plays in Clouds.
clouds_music_ids: dict[int, str] = {
		394385544: "mus_q105_dollhouse_394385544",
		643987762: "mus_q105_dollhouse_643987762",
		}

#: Music that plays in the flashbacks during the quest *Somewhat Damaged*.
somewhat_damaged_flashback_ids: dict[int, str] = {
		180066973: "mus_q305_brooklyn_vinyl_01_180066973",
		22001095: "mus_q305_brooklyn_party_22001095",
		}

#: Music that plays in the flashbacks during sex cutscenes.
sex_cutscene_music_ids: dict[int, str] = {
		415608881: "mus_generic_sexcs_01_415608881",  # Joytoy fire?
		851644048: "mus_sq030_judy_csex_851644048",  # Guitar and drums and humming
		176338671: "mus_stout_sexcs_01_176338671",  # Hole In The Sun (excerpt)
		}

#: Music that plays during the *I Really Wanna Stay At Your House* quest.
hangout_music_ids: dict[int, str] = {}

#: Music that plays during the *The Beast In Me* quests.
race_music_ids: dict[int, str] = {}

#: Unknown
mus_ow_phone_ids: dict[int, str] = {}

#: Music that plays in Dark Matter during the *Off the Leash* quest.
dark_matter_music_ids: dict[int, str] = {}

#:
johnny_kerry_music_ids: dict[int, str] = {}

for station, station_data in radio_stations.items():
	for track in station_data:
		if not track.other_uses:
			continue
		for wem_name, use in track.other_uses.items():
			if use == HANGOUT_QUEST:
				hangout_music_ids[wem_name] = f"{use}_{wem_name}"
			elif use == RACES:
				race_music_ids[wem_name] = f"{use}_{wem_name}"
			elif use == MUS_OW_PHONE:
				mus_ow_phone_ids[wem_name] = f"{use}_{wem_name}"
			elif use in {SOMEWHAT_DAMAGED_FLASHBACK, SOMEWHAT_DAMAGED_FLASHBACK_REED}:
				somewhat_damaged_flashback_ids[wem_name] = f"{use}_{wem_name}"
			elif use == CLOUDS_MUSIC:
				clouds_music_ids[wem_name] = f"{use}_{wem_name}"
			elif use == DARK_MATTER:
				dark_matter_music_ids[wem_name] = f"{use}_{wem_name}"
			elif use == JOHNNY_KERRY:
				johnny_kerry_music_ids[wem_name] = f"{use}_{wem_name}"
			elif use in {"mus_sq027_panam_csex_01_pre_sex"}:
				sex_cutscene_music_ids[wem_name] = f"{use}_{wem_name}"
			elif use in {
					"mus_ep1_credits",
					"mus_q303_restaurant_01",
					"mus_mq301_car_cmb_source",
					"mus_q001_v_apartment_chippin_in",
					"mus_q115_chippin_in",
					"mus_e3demo_end",
					"mus_q110_chippinin",
					"mus_q204_js_apartment_vinyl_01, during New Dawn Fades",
					"mus_q108_concert_glitch",
					"mus_sq011_totentanz_01",
					"Phantom Liberty Black Sapphire performance",
					"Parade and Wako's briefing",
					"Konpeki Plaza Braindance",
					"mus_q105_fingers_01_source",
					"mus_q000_corpo_standoff_source"
					}:
				pass  # TODO
			else:
				raise NotImplementedError(use)


class SceneAudioData(Protocol):
	"""
	Protocol for classes such as :class:`~.AdvertData` and :class:`~.DJData`.
	"""

	@property
	def audio_filename_prefix(self) -> str: ...

	@property
	def scene_file(self) -> str: ...

	@property
	def general_audio(self) -> bool: ...

	@overload
	def __getitem__(self, __key: SupportsIndex) -> str | bool: ...

	@overload
	def __getitem__(self, __key: "slice[Any, Any, Any]") -> tuple[str | bool, ...]: ...

	def __iter__(self) -> Iterator: ...



#!/usr/bin/env python3
#
#  work.py
"""
Classes to represent datatypes within CR2W/W2RC files (prefixed ``work``).
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import sys
from dataclasses import dataclass, field

# this package
from cp2077_extractor.cr2w import enums
from cp2077_extractor.cr2w.datatypes.anim import animAnimSet, animAnimSetup, animRig
from cp2077_extractor.cr2w.datatypes.base import Chunk, redTagList
from cp2077_extractor.cr2w.datatypes.ent import entEntityTemplate
from cp2077_extractor.cr2w.datatypes.game import gameScriptableComponent

__all__ = [
		"WorkspotEntryData",
		"WorkspotMapData",
		"WorkspotMapperComponent",
		"workIEntry",
		"workIWorkspotItemAction",
		"workTransitionAnim",
		"workWorkEntryId",
		"workWorkspotAnimsetEntry",
		"workWorkspotGlobalProp",
		"workWorkspotItemOverride",
		"workWorkspotItemOverrideItemOverride",
		"workWorkspotItemOverridePropOverride",
		"workWorkspotTree",
		]


@dataclass
class workWorkEntryId(Chunk):
	id: int = sys.maxsize


@dataclass
class workWorkspotItemOverridePropOverride(Chunk):
	prev_item_id: str = ''
	new_item_id: str = ''


@dataclass
class workWorkspotItemOverrideItemOverride(Chunk):
	prev_item_id: int = 0
	new_item_id: int = 0


@dataclass
class workWorkspotItemOverride(Chunk):
	prop_overrides: list[workWorkspotItemOverridePropOverride] = field(default_factory=list)
	item_overrides: list[workWorkspotItemOverrideItemOverride] = field(default_factory=list)


@dataclass
class WorkspotEntryData(Chunk):
	workspot_ref: str = ''
	is_enabled: bool = False
	is_available: bool = False


@dataclass
class WorkspotMapData(Chunk):
	action: enums.gamedataWorkspotActionType = enums.gamedataWorkspotActionType.DeviceInvestigation
	workspots: list[WorkspotEntryData] = field(default_factory=list)


@dataclass
class WorkspotMapperComponent(gameScriptableComponent):
	workspots_map: list[WorkspotMapData] = field(default_factory=list)


@dataclass
class workIWorkspotItemAction(Chunk):
	pass


@dataclass
class workWorkspotGlobalProp(Chunk):
	id: str = ''
	bone_name: str = ''
	prop: entEntityTemplate = field(default_factory=entEntityTemplate)  # CResourceAsyncReference


@dataclass
class workIEntry(Chunk):
	id: workWorkEntryId = field(default_factory=workWorkEntryId)
	flags: int = 0


@dataclass
class workWorkspotAnimsetEntry(Chunk):
	rig: animRig = field(default_factory=animRig)  # TODO: CResourceAsyncReference
	animations: animAnimSetup = field(default_factory=animAnimSetup)
	loading_handles: list[animAnimSet] = field(default_factory=list)  # TODO: CResourceReference


@dataclass
class workTransitionAnim(Chunk):
	idle_a: str = ''
	idle_b: str = ''
	transition_ato_b: str = ''
	transition_bto_a: str = ''


@dataclass
class workWorkspotTree(Chunk):
	workspot_rig: animRig = field(default_factory=animRig)  # TODO: CResourceAsyncReference
	global_props: list[workWorkspotGlobalProp] = field(default_factory=list)
	props_play_sync_anim: bool = False
	root_entry: workIEntry = field(default_factory=workIEntry)
	id_counter: int = 0
	dont_inject_workspot_graph: bool = False
	anim_graph_slot_name: str = "WORKSPOT"
	auto_transition_blend_time: float = 1.0
	initial_actions: list[workIWorkspotItemAction] = field(default_factory=list)
	initial_can_use_exits: bool = True
	blend_out_time: float = 0.0
	final_animsets: list[workWorkspotAnimsetEntry] = field(default_factory=list)
	tags: redTagList = field(default_factory=redTagList)
	items_policy: enums.workWorkspotItemPolicy = enums.workWorkspotItemPolicy.ItemPolicy_SpawnItemOnIdleChange | enums.workWorkspotItemPolicy.ItemPolicy_DespawnItemOnIdleChange | enums.workWorkspotItemPolicy.ItemPolicy_DespawnItemOnReaction  # TODO: CBitField
	censorship_flags: enums.CensorshipFlags = enums.CensorshipFlags.Censor_Nudity  # TODO: CBitField
	custom_transition_anims: list[workTransitionAnim] = field(default_factory=list)
	inertialization_duration_enter: float = 0.5
	inertialization_duration_exit_natural: float = 0.5
	inertialization_duration_exit_forced: float = 0.2
	use_time_limit_for_sequences: bool = False
	freze_at_the_last_frame_use_with_caution: bool = False
	sequences_time_limit: float = 1.0
	snap_to_terrain: bool = False
	unmount_body_carry: bool = True
	status_effect_id: int = 0
	whitelist_visual_tags: redTagList = field(default_factory=redTagList)
	blacklist_visual_tags: redTagList = field(default_factory=redTagList)



#!/usr/bin/env python3
#
#  utils.py
"""
General utility functions.
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import random
from collections import deque
from io import BytesIO
from typing import Deque, Generic, TypeVar

# 3rd party
import lameenc  # type: ignore[import-not-found]
import regex as re  # type: ignore[import-untyped]
from domdf_python_tools.paths import PathPlus
from miniaudio import SoundFileInfo, vorbis_get_info, vorbis_read  # type: ignore[import-untyped]
from mutagen.id3 import ID3, TLEN
from wem2ogg import wem_to_ogg

__all__ = ["InfiniteList", "StringReader", "to_snake_case", "transcode_file"]

try:
	# 3rd party
	from kraken_decompressor import decompress
except ImportError:

	def decompress(src: bytes, dst_len: int) -> bytes:
		raise NotImplementedError(
				"Kraken decompression unavailable ('kraken-decompressor' not installed or unsupported platform)"
				)


def transcode_file(
		wem_filename: PathPlus,
		mp3_filename: PathPlus,
		length_range: tuple[int, int] | None = None,
		) -> None:
	"""
	Transcode a WWise ``.wem`` file to mp3 at 256kbps.

	Requires ``ffmpeg`` to be installed.

	:param wem_filename:
	:param mp3_filename:
	:param length_range: Files with durations in seconds outside this range will be skipped.
	"""

	# TODO: see how vgmstream gets length; probably in file header

	# print(wem_filename, "->", mp3_filename)
	ogg_data = wem_to_ogg(wem_filename.read_bytes())
	ogg_info: SoundFileInfo = vorbis_get_info(ogg_data)
	# print("nchannels =", ogg_info.nchannels)
	# print("sample_rate =", ogg_info.sample_rate)
	# print("sample_width =", ogg_info.sample_width)
	# print("num_frames =", ogg_info.num_frames)
	# print("duration =", ogg_info.duration)  # Seconds
	# print("sub_format =", ogg_info.sub_format)

	length = ogg_info.duration
	if not length_range or (length_range[1] >= length >= length_range[0]):

		pcm_data = bytes(vorbis_read(data=ogg_data).samples)

		encoder = lameenc.Encoder()
		encoder.set_bit_rate(256)
		# encoder.set_in_sample_rate(sample_rate)
		# encoder.set_channels(2)
		encoder.set_in_sample_rate(ogg_info.sample_rate)
		encoder.set_channels(ogg_info.nchannels)
		encoder.set_quality(2)  # 2-highest, 7-fastest
		mp3_data = encoder.encode(pcm_data)
		mp3_data += encoder.flush()  # Flush when finished encoding the entire stream

		tags = ID3()
		tags.add(TLEN(encoding=0, data=length * 1000))
		data = tags._prepare_data(BytesIO(mp3_data), 0, 0, 4, '/', None)

		mp3_filename.write_bytes(data + mp3_data)

	# else:
	# 	print("Skip ogg; too short or too long")


_T = TypeVar("_T")


class InfiniteList(Generic[_T]):
	"""
	List-like object that refills with a random order once empty.
	"""

	_items: list[_T]
	_recent: Deque[_T]
	_working_items: list[_T]

	def __init__(self, items: list[_T]) -> None:
		self._items = items[:]
		if items:
			# self._recent = deque(maxlen=min(len(items) - 1, 5))
			self._recent = deque(maxlen=min(len(items) - 1, 5))
		else:
			self._recent = deque()

		self.repopulate()

	def repopulate(self) -> None:
		"""
		Repopulate the list with a new random order, avoiding recent items occuring soon.
		"""

		# print("Starting repopulate")
		self._working_items = []
		remaining_items = self._items[:]
		while remaining_items:
			# item = random.choice(remaining_items)
			# if len(self._working_items) < self._recent.maxlen:
			# 	if item in self._recent:
			# 		continue
			choices = []
			for item in remaining_items:
				assert self._recent.maxlen is not None
				if len(self._working_items) < self._recent.maxlen:
					if item in self._recent:
						continue
				choices.append(item)
			if not choices:
				choices = remaining_items
			item = random.choice(choices)
			self._working_items.append(item)
			remaining_items.remove(item)

		self._working_items.reverse()

	def pop(self) -> _T:
		"""
		Get the next item from the back of the list.
		"""

		if not self._working_items:
			self.repopulate()

		item = self._working_items.pop()
		self._recent.append(item)

		return item


_case_boundary_re = re.compile("(\\p{Ll})(\\p{Lu})")
_single_letters_re = re.compile("(\\p{Lu}|\\p{N})(\\p{Lu})(\\p{Ll})")


def to_snake_case(value: str) -> str:
	"""
	Convert the given string into ``snake_case``.

	:param value:
	"""

	# Matches VSCode behaviour
	case_boundary = _case_boundary_re.findall(value)
	single_letters = _single_letters_re.findall(value)
	if not case_boundary and not single_letters:
		return value.lower()
	value = _case_boundary_re.sub(r"\1_\2", value)
	value = _case_boundary_re.sub(r"\1_\2\3", value)
	return value.lower()


class StringReader(BytesIO):
	"""
	Reader for REDengine sized strings.
	"""

	vlq_value_mask = 0b01111111
	vlq_continuation = 0b10000000

	def parse_string_and_size(self) -> tuple[int, str]:
		"""
		Parse a length-prefixed string (as bytes) to a Python string.

		:param value: The string with a VLQ i32 length prefix.

		:returns: Tuple of length prefix and the string.
		"""

		size_prefix = self.read_vlq_int32()

		# The string length is the absolute value of the size prefix
		string_length = abs(size_prefix)

		if not string_length:
			return size_prefix, ''

		# Sign bit indicates whether UTF-16 (0) or UTF-8 (1)
		if size_prefix > 0:
			encoding = "UTF-16"
		else:
			encoding = "UTF-8"

		return size_prefix, self.read(string_length).decode(encoding)

	def parse_string(self) -> str:
		"""
		Parse a length-prefixed string (as bytes) to a Python string.

		:param value: The string with a VLQ i32 length prefix.
		"""

		return self.parse_string_and_size()[1]

	def read_vlq_int32(self) -> int:
		"""
		Parse modified 32 bit VLQ to int.

		The first bit is the sign bit, the 2nd bit tells whether there are more octets to read,
		and the next 6 bytes are the least significant bits of the number data.
		Remaining octets are 1+7 continuation and data.
		"""

		b = self.read(1)[0]
		is_negative = bool(b & 0b10000000)

		# Take the initial value from the lower 6 bits
		value = b & 0b00111111

		# Is the value larger than 6 bits?
		if (b & 0b01000000):  # The first octet stores the continuation flag in the 6th bit
			b = self.read(1)[0]
			# Mask and add the next 7 bits
			value |= (b & self.vlq_value_mask) << 6

			# Is the value larger than 13 bits?
			if (b & self.vlq_continuation):
				b = self.read(1)[0]
				value |= (b & self.vlq_value_mask) << 13

				# Is the value larger than 20 bits?
				if (b & self.vlq_continuation):
					b = self.read(1)[0]
					value |= (b & self.vlq_value_mask) << 20

					# Is the value larger than 27 bits?
					if (b & self.vlq_continuation):
						b = self.read(1)[0]
						value |= (b & self.vlq_value_mask) << 27

						# Is the value larger than 34 bits? That seems bad
						if (b & self.vlq_continuation):
							raise ValueError("Continuation bit set on 5th byte")

		if is_negative:
			return -value
		else:
			return value
